import json
from datetime import timedelta

from django.contrib.auth import get_user_model
from django.urls import reverse

from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...threads.models import Post, Thread
from ...threads.test import post_thread
from ..activepostersranking import build_active_posters_ranking
from ..models import Ban, DeletedUser, Rank
from ..test import AuthenticatedUserTestCase, create_test_user

User = get_user_model()


class ActivePostersListTests(AuthenticatedUserTestCase):
    """tests for active posters list (GET /users/?list=active)"""

    def setUp(self):
        super().setUp()

        self.link = "/api/users/?list=active"

        self.category = Category.objects.all_categories()[:1][0]
        self.category.labels = []

    def test_empty_list(self):
        """empty list is served"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, self.user.username)

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, self.user.username)

    def test_filled_list(self):
        """filled list is served"""
        post_thread(self.category, poster=self.user)
        self.user.posts = 1
        self.user.save()

        build_active_posters_ranking()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.user.username)
        self.assertContains(response, '"is_online":true')
        self.assertContains(response, '"is_offline":false')

        self.logout_user()
        build_active_posters_ranking()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.user.username)
        self.assertContains(response, '"is_online":false')
        self.assertContains(response, '"is_offline":true')


class FollowersListTests(AuthenticatedUserTestCase):
    """tests for generic list (GET /users/) filtered by followers"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/followers/"

    def test_nonexistent_user(self):
        """list for non-existing user returns 404"""
        response = self.client.get(self.link % 31242)
        self.assertEqual(response.status_code, 404)

    def test_empty_list(self):
        """user without followers returns 200"""
        response = self.client.get(self.link % self.user.pk)
        self.assertEqual(response.status_code, 200)

    def test_filled_list(self):
        """user with followers returns 200"""
        follower = create_test_user("TestFollower", "test@follower.com")
        self.user.followed_by.add(follower)

        response = self.client.get(self.link % self.user.pk)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, follower.username)

    def test_filled_list_search(self):
        """followers list is searchable"""
        follower = create_test_user("TestFollower", "test@follower.com")
        self.user.followed_by.add(follower)

        api_link = self.link % self.user.pk

        response = self.client.get("%s?search=%s" % (api_link, "test"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, follower.username)


class FollowsListTests(AuthenticatedUserTestCase):
    """tests for generic list (GET /users/) filtered by follows"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/follows/"

    def test_nonexistent_user(self):
        """list for non-existing user returns 404"""
        response = self.client.get(self.link % 1321)
        self.assertEqual(response.status_code, 404)

    def test_empty_list(self):
        """user without follows returns 200"""
        response = self.client.get(self.link % self.user.pk)
        self.assertEqual(response.status_code, 200)

    def test_filled_list(self):
        """user with follows returns 200"""
        follower = create_test_user("TestFollower", "test@follower.com")
        self.user.follows.add(follower)

        response = self.client.get(self.link % self.user.pk)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, follower.username)

    def test_filled_list_search(self):
        """follows list is searchable"""
        follower = create_test_user("TestFollower", "test@follower.com")
        self.user.follows.add(follower)

        api_link = self.link % self.user.pk

        response = self.client.get("%s?search=%s" % (api_link, "test"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, follower.username)


class RankListTests(AuthenticatedUserTestCase):
    """tests for generic list (GET /users/) filtered by rank"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/?rank=%s"

    def test_nonexistent_rank(self):
        """list for non-existing rank returns 404"""
        response = self.client.get(self.link % 1421)
        self.assertEqual(response.status_code, 404)

    def test_empty_list(self):
        """tab rank without members returns 200"""
        test_rank = Rank.objects.create(name="Test rank", slug="test-rank", is_tab=True)

        response = self.client.get(self.link % test_rank.pk)
        self.assertEqual(response.status_code, 200)

    def test_disabled_list(self):
        """non-tab rank returns 404"""
        self.user.rank.is_tab = False
        self.user.rank.save()

        response = self.client.get(self.link % self.user.rank.pk)
        self.assertEqual(response.status_code, 404)

    def test_list_search(self):
        """rank list is not searchable"""
        api_link = self.link % self.user.rank.pk

        response = self.client.get("%s&name=%s" % (api_link, "test"))
        self.assertEqual(response.status_code, 404)

    def test_filled_list(self):
        """tab rank with members return 200"""
        self.user.rank.is_tab = True
        self.user.rank.save()

        response = self.client.get(self.link % self.user.rank.pk)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.user.username)

    def test_disabled_users(self):
        """api follows disabled users visibility"""
        rank = Rank.objects.create(name="Test rank", slug="test-rank", is_tab=True)
        user = create_test_user(
            "DisabledUser", "disabled@example.com", rank=rank, is_active=False
        )

        response = self.client.get(self.link % rank.pk)
        self.assertNotContains(response, user.get_absolute_url())

        # api shows disabled accounts to staff
        self.user.is_staff = True
        self.user.save()

        response = self.client.get(self.link % rank.pk)
        self.assertContains(response, user.get_absolute_url())


class SearchNamesListTests(AuthenticatedUserTestCase):
    """tests for generic list (GET /users/) filtered by username disallowing searches"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/?&name="

    def test_empty_list(self):
        """empty list returns 404"""
        response = self.client.get(self.link + "this-user-is-fake")
        self.assertEqual(response.status_code, 404)

    def test_filled_list(self):
        """results list returns 404"""
        response = self.client.get(self.link + self.user.slug)
        self.assertEqual(response.status_code, 404)


class UserRetrieveTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "otheruser@example.com")
        self.link = reverse("misago:api:user-detail", kwargs={"pk": self.other_user.pk})

    def test_get_user(self):
        """api user retrieve endpoint has no showstoppers"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

    def test_disabled_user(self):
        """api user retrieve handles disabled users"""
        self.user.is_staff = False
        self.user.save()

        self.other_user.is_active = False
        self.other_user.save()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 404)

        self.user.is_staff = True
        self.user.save()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)


class UserFollowTests(AuthenticatedUserTestCase):
    """tests for user follow RPC (POST to /api/users/1/follow/)"""

    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "otheruser@example.com")
        self.link = "/api/users/%s/follow/" % self.other_user.pk

    def test_follow_unauthenticated(self):
        """you have to sign in to follow users"""
        self.logout_user()

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    def test_follow_myself(self):
        """you can't follow yourself"""
        response = self.client.post("/api/users/%s/follow/" % self.user.pk)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't add yourself to followed."}
        )

    @patch_user_acl({"can_follow_users": 0})
    def test_cant_follow(self):
        """no permission to follow users"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't follow other users."})

    def test_follow(self):
        """follow and unfollow other user"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 200)

        self.user.refresh_from_db()
        self.assertEqual(self.user.followers, 0)
        self.assertEqual(self.user.following, 1)
        self.assertEqual(self.user.follows.count(), 1)
        self.assertEqual(self.user.followed_by.count(), 0)

        self.other_user.refresh_from_db()
        self.assertEqual(self.other_user.followers, 1)
        self.assertEqual(self.other_user.following, 0)
        self.assertEqual(self.other_user.follows.count(), 0)
        self.assertEqual(self.other_user.followed_by.count(), 1)

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 200)

        self.user.refresh_from_db()
        self.assertEqual(self.user.followers, 0)
        self.assertEqual(self.user.following, 0)
        self.assertEqual(self.user.follows.count(), 0)
        self.assertEqual(self.user.followed_by.count(), 0)

        self.other_user.refresh_from_db()
        self.assertEqual(self.other_user.followers, 0)
        self.assertEqual(self.other_user.following, 0)
        self.assertEqual(self.other_user.follows.count(), 0)
        self.assertEqual(self.other_user.followed_by.count(), 0)


class UserBanTests(AuthenticatedUserTestCase):
    """tests for ban endpoint (GET to /api/users/1/ban/)"""

    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "otheruser@example.com")
        self.link = "/api/users/%s/ban/" % self.other_user.pk

    @patch_user_acl({"can_see_ban_details": 0})
    def test_no_permission(self):
        """user has no permission to access ban"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't see users bans details."}
        )

    @patch_user_acl({"can_see_ban_details": 1})
    def test_no_ban(self):
        """api returns empty json"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {})

    @patch_user_acl({"can_see_ban_details": 1})
    def test_ban_details(self):
        """api returns ban json"""
        Ban.objects.create(
            check_type=Ban.USERNAME,
            banned_value=self.other_user.username,
            user_message="Nope!",
        )

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        ban_json = response.json()
        self.assertEqual(ban_json["user_message"]["plain"], "Nope!")
        self.assertEqual(ban_json["user_message"]["html"], "<p>Nope!</p>")


class UserDeleteOwnAccountTests(AuthenticatedUserTestCase):
    """
    tests for user request own account delete RPC
    (POST to /api/users/1/delete-own-account/)
    """

    def setUp(self):
        super().setUp()
        self.api_link = "/api/users/%s/delete-own-account/" % self.user.pk

    @override_dynamic_settings(allow_delete_own_account=False)
    def test_delete_own_account_feature_disabled(self):
        """
        raises 403 error when attempting to delete own account but feature is disabled
        """
        response = self.client.post(self.api_link, {"password": self.USER_PASSWORD})

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't delete your account."})

        self.reload_user()
        self.assertTrue(self.user.is_active)
        self.assertFalse(self.user.is_deleting_account)

    @override_dynamic_settings(allow_delete_own_account=True)
    def test_delete_own_account_is_staff(self):
        """raises 403 error when attempting to delete own account as admin"""
        self.user.is_staff = True
        self.user.save()

        response = self.client.post(self.api_link, {"password": self.USER_PASSWORD})
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "You can't delete your account because you are an administrator."
                )
            },
        )

        self.reload_user()
        self.assertTrue(self.user.is_active)
        self.assertFalse(self.user.is_deleting_account)

    @override_dynamic_settings(allow_delete_own_account=True)
    def test_delete_own_account_is_superuser(self):
        """raises 403 error when attempting to delete own account as superadmin"""
        self.user.is_superuser = True
        self.user.save()

        response = self.client.post(self.api_link, {"password": self.USER_PASSWORD})
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": (
                    "You can't delete your account because you are an administrator."
                )
            },
        )

        self.reload_user()
        self.assertTrue(self.user.is_active)
        self.assertFalse(self.user.is_deleting_account)

    @override_dynamic_settings(allow_delete_own_account=True)
    def test_delete_own_account_invalid_password(self):
        """
        raises 400 error when attempting to delete own account with invalid password
        """
        response = self.client.post(self.api_link, {"password": "hello"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"password": ["Entered password is invalid."]}
        )

        self.reload_user()
        self.assertTrue(self.user.is_active)
        self.assertFalse(self.user.is_deleting_account)

    @override_dynamic_settings(allow_delete_own_account=True)
    def test_delete_own_account(self):
        """deactivates account and marks it for deletion"""
        response = self.client.post(self.api_link, {"password": self.USER_PASSWORD})
        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertFalse(self.user.is_active)
        self.assertTrue(self.user.is_deleting_account)


@override_dynamic_settings(
    allow_delete_own_account=True,
    enable_oauth2_client=True,
    oauth2_provider="Lorem",
)
def test_delete_own_account_api_returns_403_if_oauth_is_enabled(
    user, user_password, user_client
):
    response = user_client.post(
        reverse("misago:api:user-delete-own-account", kwargs={"pk": user.pk}),
        {
            "password": user_password,
        },
    )
    assert response.status_code == 403


class UserDeleteTests(AuthenticatedUserTestCase):
    """tests for user delete RPC (POST to /api/users/1/delete/)"""

    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "otheruser@example.com")
        self.link = "/api/users/%s/delete/" % self.other_user.pk

        self.threads = Thread.objects.count()
        self.posts = Post.objects.count()

        self.category = Category.objects.all_categories()[:1][0]

        post_thread(self.category, poster=self.other_user)
        self.other_user.posts = 1
        self.other_user.threads = 1
        self.other_user.save()

    @patch_user_acl(
        {"can_delete_users_newer_than": 0, "can_delete_users_with_less_posts_than": 0}
    )
    def test_delete_no_permission(self):
        """raises 403 error when no permission to delete"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't delete users."})

    @patch_user_acl(
        {"can_delete_users_newer_than": 0, "can_delete_users_with_less_posts_than": 5}
    )
    def test_delete_too_many_posts(self):
        """raises 403 error when user has too many posts"""
        self.other_user.posts = 6
        self.other_user.save()

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete users that made more than 5 posts."},
        )

    @patch_user_acl(
        {"can_delete_users_newer_than": 5, "can_delete_users_with_less_posts_than": 0}
    )
    def test_delete_too_old_member(self):
        """raises 403 error when user is too old"""
        self.other_user.joined_on -= timedelta(days=6)
        self.other_user.save()

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete users that are members for more than 5 days."},
        )

    @patch_user_acl(
        {"can_delete_users_newer_than": 10, "can_delete_users_with_less_posts_than": 10}
    )
    def test_delete_self(self):
        """raises 403 error when attempting to delete oneself"""
        response = self.client.post("/api/users/%s/delete/" % self.user.pk)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't delete your account."})

    @patch_user_acl(
        {"can_delete_users_newer_than": 10, "can_delete_users_with_less_posts_than": 10}
    )
    def test_delete_admin(self):
        """raises 403 error when attempting to delete admin"""
        self.other_user.is_staff = True
        self.other_user.save()

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "Administrators can't be deleted."}
        )

    @patch_user_acl(
        {"can_delete_users_newer_than": 10, "can_delete_users_with_less_posts_than": 10}
    )
    def test_delete_superadmin(self):
        """raises 403 error when attempting to delete superadmin"""
        self.other_user.is_superuser = True
        self.other_user.save()

        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "Administrators can't be deleted."}
        )

    @patch_user_acl(
        {"can_delete_users_newer_than": 10, "can_delete_users_with_less_posts_than": 10}
    )
    def test_delete_with_content(self):
        """returns 200 and deletes user with content"""
        response = self.client.post(
            self.link,
            json.dumps({"with_content": True}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        with self.assertRaises(User.DoesNotExist):
            self.other_user.refresh_from_db()

        self.assertEqual(Thread.objects.count(), self.threads)
        self.assertEqual(Post.objects.count(), self.posts)

    @patch_user_acl(
        {"can_delete_users_newer_than": 10, "can_delete_users_with_less_posts_than": 10}
    )
    def test_delete_without_content(self):
        """returns 200 and deletes user without content"""
        response = self.client.post(
            self.link,
            json.dumps({"with_content": False}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        with self.assertRaises(User.DoesNotExist):
            self.other_user.refresh_from_db()

        self.assertEqual(Thread.objects.count(), self.threads + 1)
        self.assertEqual(Post.objects.count(), self.posts + 2)

    @patch_user_acl(
        {"can_delete_users_newer_than": 10, "can_delete_users_with_less_posts_than": 10}
    )
    def test_deleting_user_using_api_records_deletion_by_staff(self):
        response = self.client.post(
            self.link,
            json.dumps({"with_content": False}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_STAFF)



from . import useracl


def user_acl_middleware(get_response):
    """Sets request.user_acl attribute with dict containing current user acl."""

    def middleware(request):
        request.user_acl = useracl.get_user_acl(request.user, request.cache_versions)
        return get_response(request)

    return middleware



import requests
from django.core.exceptions import ValidationError
from django.utils.translation import pgettext


def recaptcha_test(request):
    r = requests.post(
        "https://www.google.com/recaptcha/api/siteverify",
        data={
            "secret": request.settings.recaptcha_secret_key,
            "response": request.data.get("captcha"),
            "remoteip": request.user_ip,
        },
    )

    if r.status_code == 200:
        response_json = r.json()
        if not response_json.get("success"):
            raise ValidationError(pgettext("recaptcha check", "Please try again."))
    else:
        raise ValidationError(
            pgettext("recaptcha check", "Failed to contact reCAPTCHA API.")
        )


def qacaptcha_test(request):
    answer = request.data.get("captcha", "").lower().strip()
    valid_answers = get_valid_qacaptcha_answers(request.settings)
    if answer not in valid_answers:
        raise ValidationError(
            pgettext("qacaptcha check", "Entered answer is incorrect.")
        )


def get_valid_qacaptcha_answers(settings):
    valid_answers = [i.strip() for i in settings.qa_answers.lower().splitlines()]
    return filter(len, valid_answers)


def nocaptcha_test(request):
    return  # no captcha means no validation


CAPTCHA_TESTS = {"re": recaptcha_test, "qa": qacaptcha_test, "no": nocaptcha_test}


def test_request(request):
    CAPTCHA_TESTS[request.settings.captcha_type](request)



from ..parser import parse


def test_hr_bbcode_is_replaced_if_its_alone_in_paragraph(request_mock, user, snapshot):
    text = """
Lorem ipsum dolor met.
[hr]
Sit amet elit.
"""
    result = parse(text, request_mock, user)
    assert snapshot == result["parsed_text"]


def test_hr_bbcode_is_skipped_if_its_part_of_paragraph(request_mock, user, snapshot):
    text = "Lorem ipsum[hr]dolor met."
    result = parse(text, request_mock, user)
    assert result["parsed_text"] == "<p>Lorem ipsum[hr]dolor met.</p>"



from functools import wraps

from django.core.exceptions import PermissionDenied
from django.utils.translation import pgettext


def require_auth(f):
    @wraps(f)
    def auth_only_view(request, *args, **kwargs):
        if request.user.is_anonymous:
            if request.method in ("GET", "HEAD", "OPTIONS"):
                raise PermissionDenied(
                    pgettext("apiv2", "You have to sign in to access this page.")
                )
            raise PermissionDenied(
                pgettext("apiv2", "You have to sign in to perform this action.")
            )

        return f(request, *args, **kwargs)

    return auth_only_view



from django.contrib.auth.models import Permission
from django.test import override_settings
from django.urls import reverse
from django.utils import formats

from ....admin.test import AdminTestCase
from ...test import create_test_user
from ..djangoadmin import UserAdminModel


@override_settings(ROOT_URLCONF="misago.core.testproject.urls")
class TestDjangoAdminUserForm(AdminTestCase):
    def setUp(self):
        super().setUp()
        self.test_user = create_test_user("Other_User", "otheruser@example.com")
        self.edit_test_user_in_django_url = reverse(
            "admin:misago_users_user_change", args=[self.test_user.pk]
        )
        self.edit_test_user_in_misago_url = reverse(
            "misago:admin:users:edit", args=[self.test_user.pk]
        )

    def test_user_edit_view_content(self):
        """basic data about a user is present in Django admin"""
        response = self.client.get(self.edit_test_user_in_django_url)

        self.assertEqual(response.status_code, 200)

        self.assertContains(response, self.test_user.username)
        self.assertContains(response, self.test_user.email)
        self.assertContains(response, self.test_user.rank)

        last_login_date = formats.date_format(self.test_user.last_login)
        self.assertContains(response, last_login_date)

        register_date = formats.date_format(self.test_user.joined_on)
        self.assertContains(response, register_date)

        user_permissions = Permission.objects.all()
        for permission in user_permissions:
            self.assertContains(response, permission)

    def test_user_edit_view_post(self):
        """user permissions are editable through Django admin"""
        perms_all = Permission.objects.all()
        perms_all_pks = []
        for perm in perms_all:
            perms_all_pks.append(perm.pk)

        response = self.client.post(
            self.edit_test_user_in_django_url, data={"user_permissions": perms_all_pks}
        )
        self.assertEqual(response.status_code, 302)

        user_perms = self.test_user.user_permissions.all()
        for perm_pk in perms_all_pks:
            is_user_has_perm = user_perms.filter(pk=perm_pk).exists()
            self.assertTrue(is_user_has_perm)

    def test_misago_admin_url_presence_in_user_edit_view(self):
        """the url to Misago admin is present in Django admin user edit view"""
        response = self.client.get(self.edit_test_user_in_django_url)
        self.assertContains(response, self.edit_test_user_in_misago_url)
        edit_from_misago_short_desc = (
            UserAdminModel.get_edit_from_misago_url.short_description
        )
        self.assertContains(response, edit_from_misago_short_desc)

    def test_misago_admin_url_presence_in_user_list_view(self):
        """the url to Misago admin is present in Django admin user list view"""
        response = self.client.get(reverse("admin:misago_users_user_changelist"))
        self.assertContains(response, self.edit_test_user_in_misago_url)



from django.shortcuts import get_object_or_404, redirect, render

from .models import Agreement
from .utils import get_parsed_agreement_text


def legal_view(request, agreement_type):
    agreement = get_object_or_404(Agreement, type=agreement_type, is_active=True)

    if agreement.link:
        return redirect(agreement.link)

    template_name = "misago/%s.html" % agreement_type
    agreement_text = get_parsed_agreement_text(request, agreement)

    return render(
        request,
        template_name,
        {
            "title": agreement.get_final_title(),
            "link": agreement.link,
            "text": agreement_text,
        },
    )


def privacy_policy(request):
    return legal_view(request, Agreement.TYPE_PRIVACY)


def terms_of_service(request):
    return legal_view(request, Agreement.TYPE_TOS)



from time import time
from unittest.mock import Mock

import pytest
from django.test import override_settings

from ..auth import (
    TOKEN_KEY,
    UPDATED_KEY,
    authorize_admin,
    is_admin_authorized,
    remove_admin_authorization,
    update_admin_authorization,
)


@pytest.fixture
def admin_request(superuser):
    request = Mock(session={}, user=superuser)
    authorize_admin(request)
    return request


def test_authorizing_admin_updates_request_session(user):
    request = Mock(session={}, user=user)
    authorize_admin(request)
    assert request.session


def test_staff_user_can_be_authorized(staffuser):
    request = Mock(session={}, user=staffuser)
    authorize_admin(request)
    assert is_admin_authorized(request)


def test_non_staff_user_admin_authorization_is_never_valid(user):
    request = Mock(session={}, user=user)
    authorize_admin(request)
    assert not is_admin_authorized(request)


def test_anonymous_user_admin_authorization_is_never_valid(user, anonymous_user):
    request = Mock(session={}, user=user)
    authorize_admin(request)
    request.user = anonymous_user
    assert not is_admin_authorized(request)


def test_superuser_without_staff_flag_admin_authorization_is_never_valid(staffuser):
    request = Mock(session={}, user=staffuser)
    authorize_admin(request)
    request.user.is_staff = False
    assert not is_admin_authorized(request)


def test_admin_authorization_is_invalidated_by_user_pk_change(admin_request, superuser):
    admin_request.user.pk = superuser.pk + 1
    assert not is_admin_authorized(admin_request)


def test_admin_authorization_is_invalidated_by_user_email_change(admin_request):
    admin_request.user.email = "changed@example.com"
    assert not is_admin_authorized(admin_request)


def test_admin_authorization_is_invalidated_by_user_password_change(admin_request):
    admin_request.user.set_password("changed-password")
    assert not is_admin_authorized(admin_request)


def test_admin_authorization_is_invalidated_by_secret_key_change(admin_request):
    with override_settings(SECRET_KEY="changed-secret-key"):
        assert not is_admin_authorized(admin_request)


def test_admin_authorization_is_invalidated_by_token_change(admin_request):
    admin_request.session[TOKEN_KEY] = "authorization-token-changed"
    assert not is_admin_authorized(admin_request)


@override_settings(MISAGO_ADMIN_SESSION_EXPIRATION=5)
def test_admin_authorization_is_invalidated_by_token_expiration(admin_request):
    admin_request.session[UPDATED_KEY] = time() - 5 * 60 - 1
    assert not is_admin_authorized(admin_request)


def test_updating_authorization_extends_authorization_expiration_time(admin_request):
    admin_request.session[UPDATED_KEY] = 0
    update_admin_authorization(admin_request)
    assert admin_request.session[UPDATED_KEY]


def test_updating_authorization_validates_authorization(admin_request):
    admin_request.session[UPDATED_KEY] = 0
    update_admin_authorization(admin_request)
    assert is_admin_authorized(admin_request)


def test_removing_authorization_removes_autorization_from_request_session(
    admin_request,
):
    admin_request.session[UPDATED_KEY] = 0
    remove_admin_authorization(admin_request)
    assert not admin_request.session


def test_removing_authorization_invalidates_autorization(admin_request):
    admin_request.session[UPDATED_KEY] = 0
    remove_admin_authorization(admin_request)
    assert not is_admin_authorized(admin_request)



# Generated by Django 2.2.3 on 2019-09-15 19:15

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="MenuItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("url", models.URLField()),
                (
                    "menu",
                    models.CharField(
                        choices=[
                            ("both", "Navbar and footer"),
                            ("navbar", "Navbar"),
                            ("footer", "Footer"),
                        ],
                        max_length=6,
                    ),
                ),
                ("order", models.IntegerField(default=0)),
                ("css_class", models.CharField(blank=True, max_length=255, null=True)),
                ("target_blank", models.BooleanField(default=False)),
                ("rel", models.CharField(blank=True, max_length=255, null=True)),
            ],
            options={"ordering": ("order",), "get_latest_by": "order"},
        )
    ]



# Generated by Django 1.11.9 on 2018-03-10 22:34
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("misago_threads", "0007_auto_20171008_0131"),
    ]

    operations = [
        migrations.AddField(
            model_name="thread",
            name="best_answer",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="misago_threads.Post",
            ),
        ),
        migrations.AddField(
            model_name="thread",
            name="best_answer_is_protected",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="thread",
            name="best_answer_marked_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="marked_best_answer_set",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="thread",
            name="best_answer_marked_by_name",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="thread",
            name="best_answer_marked_by_slug",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name="thread",
            name="best_answer_marked_on",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from rest_framework import serializers

from ..models import ThreadParticipant

__all__ = ["ThreadParticipantSerializer"]


class ThreadParticipantSerializer(serializers.ModelSerializer):
    id = serializers.SerializerMethodField()
    username = serializers.SerializerMethodField()
    avatars = serializers.SerializerMethodField()

    url = serializers.SerializerMethodField()

    class Meta:
        model = ThreadParticipant
        fields = ["id", "username", "avatars", "url", "is_owner"]

    def get_id(self, obj):
        return obj.user.id

    def get_username(self, obj):
        return obj.user.username

    def get_avatars(self, obj):
        return obj.user.avatars

    def get_url(self, obj):
        return obj.user.get_absolute_url()



import random
import time

from django.core.management.base import BaseCommand
from faker import Factory

from ....core.management.progressbar import show_progress
from ....users.models import Rank
from ...users import (
    get_fake_inactive_user,
    get_fake_admin_activated_user,
    get_fake_banned_user,
    get_fake_user,
)


class Command(BaseCommand):
    help = "Creates fake users for dev and testing purposes."

    def add_arguments(self, parser):
        parser.add_argument(
            "users", help="number of users to create", nargs="?", type=int, default=5
        )

    def handle(self, *args, **options):
        items_to_create = options["users"]

        fake = Factory.create()
        ranks = list(Rank.objects.all())

        message = "Creating %s fake user accounts...\n"
        self.stdout.write(message % items_to_create)

        created_count = 0
        start_time = time.time()
        show_progress(self, created_count, items_to_create)

        while created_count < items_to_create:
            rank = random.choice(ranks)
            if random.randint(0, 100) > 80:
                get_fake_inactive_user(fake, rank)
            elif random.randint(0, 100) > 90:
                get_fake_admin_activated_user(fake, rank)
            elif random.randint(0, 100) > 90:
                get_fake_banned_user(fake, rank)
            else:
                get_fake_user(fake, rank)

            created_count += 1
            show_progress(self, created_count, items_to_create, start_time)

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\n\nSuccessfully created %s fake user accounts in %s"
        self.stdout.write(message % (created_count, total_humanized))



from ..validation import filter_name


def test_filter_replaces_spaces_with_underscores(db):
    assert filter_name(None, "John Doe") == "John_Doe"


def test_filter_strips_special_characters(db):
    assert filter_name(None, "John Doe!") == "John_Doe"


def test_filter_strips_unicode_strings(db):
    assert filter_name(None, "Łóżąć Bęś") == "Lozac_Bes"


def test_filter_generates_random_name_to_avoid_empty_result(db):
    user_name = filter_name(None, "__!!!")
    assert user_name
    assert len(user_name) > 5
    assert user_name.startswith("User_")


def test_filter_generates_unique_name_to_avoid_same_names(user):
    user_name = filter_name(None, user.username)

    assert user_name
    assert user_name != user.username
    assert user_name.startswith(user.username)


def test_filter_keeps_user_name_if_its_same_as_current_one(user):
    user_name = filter_name(user, user.username)
    assert user_name == user.username


def test_filter_keeps_cleaned_user_name_if_its_same_as_current_one(user):
    user.set_username("Lozac_Bes")
    user.save()

    user_name = filter_name(user, "Łóżąć Bęś")
    assert user_name == user.username



import pytest

from ...threads.models import ThreadParticipant
from ...users.bans import ban_user
from ..models import Notification, WatchedThread
from ..tasks import notify_on_new_private_thread
from ..threads import ThreadNotifications
from ..verbs import NotificationVerb


@pytest.fixture
def notify_participant_mock(mocker):
    return mocker.patch(
        "misago.notifications.tasks.notify_participant_on_new_private_thread"
    )


def test_notify_on_new_private_thread_does_nothing_if_actor_is_not_found(
    notify_participant_mock, other_user, private_thread
):
    notify_on_new_private_thread(other_user.id + 1, private_thread.id, [other_user.id])
    notify_participant_mock.assert_not_called()


def test_notify_on_new_private_thread_skips_banned_users(
    notify_participant_mock, user, other_user, user_private_thread
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    ban_user(other_user)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])
    notify_participant_mock.assert_not_called()


def test_notify_on_new_private_thread_skips_inactive_users(
    notify_participant_mock, user, inactive_user, user_private_thread
):
    inactive_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    inactive_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    inactive_user.save()

    ThreadParticipant.objects.create(user=inactive_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [inactive_user.id])
    notify_participant_mock.assert_not_called()


def test_notify_on_new_private_thread_handles_exceptions(
    mocker, user, other_user, private_thread
):
    ThreadParticipant.objects.create(user=other_user, thread=private_thread)

    notify_participant_mock = mocker.patch(
        "misago.notifications.tasks.notify_participant_on_new_private_thread",
        side_effect=ValueError("Unknown"),
    )

    notify_on_new_private_thread(user.id, private_thread.id, [other_user.id])
    notify_participant_mock.assert_called_once()


def test_notify_on_new_private_thread_creates_participant_watched_thread(
    user, other_user, user_private_thread
):
    other_user.watch_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.watch_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    WatchedThread.objects.get(
        user=other_user,
        category=user_private_thread.category,
        thread=user_private_thread,
        send_emails=True,
    )


def test_notify_on_new_private_thread_from_followed_creates_participant_watched_thread(
    user, other_user, user_private_thread
):
    other_user.watch_new_private_threads_by_followed = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.watch_new_private_threads_by_other_users = ThreadNotifications.NONE
    other_user.save()

    other_user.follows.add(user)

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    WatchedThread.objects.get(
        user=other_user,
        category=user_private_thread.category,
        thread=user_private_thread,
        send_emails=True,
    )


def test_notify_on_new_private_thread_notifies_participant(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = ThreadNotifications.SITE_ONLY
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 0


def test_notify_on_new_private_thread_notifies_participant_with_email(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 1


def test_notify_on_new_private_thread_notifies_participant_following_actor(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.SITE_ONLY
    other_user.notify_new_private_threads_by_other_users = ThreadNotifications.NONE
    other_user.save()

    other_user.follows.add(user)

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 0


def test_notify_on_new_private_thread_notifies_participant_following_actor_with_email(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.notify_new_private_threads_by_other_users = ThreadNotifications.NONE
    other_user.save()

    other_user.follows.add(user)

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.get(user=other_user, actor=user)
    assert len(mailoutbox) == 1


def test_notify_on_new_private_thread_skips_notification_if_one_already_exists(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    Notification.objects.create(
        user=other_user,
        verb=NotificationVerb.INVITED,
        category=user_private_thread.category,
        thread=user_private_thread,
        post=user_private_thread.first_post,
        is_read=False,
    )

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 0

    Notification.objects.count() == 1
    assert len(mailoutbox) == 0


def test_notify_on_new_private_thread_notifies_participant_if_old_notification_is_read(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    Notification.objects.create(
        user=other_user,
        verb=NotificationVerb.INVITED,
        category=user_private_thread.category,
        thread=user_private_thread,
        post=user_private_thread.first_post,
        is_read=True,
    )

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 1

    Notification.objects.count() == 2
    assert len(mailoutbox) == 1


def test_notify_on_new_private_thread_skips_participant_if_they_have_no_permission(
    mocker, user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    ThreadParticipant.objects.create(user=other_user, thread=user_private_thread)

    can_use_private_threads_mock = mocker.patch(
        "misago.notifications.threads.can_use_private_threads",
        return_value=False,
    )

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 0

    assert not Notification.objects.exists()
    assert len(mailoutbox) == 0

    can_use_private_threads_mock.assert_called_once()


def test_notify_on_new_private_thread_skips_user_not_participating(
    user, other_user, user_private_thread, mailoutbox
):
    other_user.notify_new_private_threads_by_followed = ThreadNotifications.NONE
    other_user.notify_new_private_threads_by_other_users = (
        ThreadNotifications.SITE_AND_EMAIL
    )
    other_user.save()

    notify_on_new_private_thread(user.id, user_private_thread.id, [other_user.id])

    other_user.refresh_from_db()
    assert other_user.unread_notifications == 0

    assert not Notification.objects.exists()
    assert len(mailoutbox) == 0



from ..datadownloads import request_user_data_download
from ..test import AuthenticatedUserTestCase


class UserDataDownloadsApiTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/data-downloads/" % self.user.pk

    def test_get_other_user_exports_anonymous(self):
        """requests to api fails if user is anonymous"""
        self.logout_user()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have to sign in to perform this action."}
        )

    def test_get_other_user_exports(self):
        """requests to api fails if user tries to access other user"""
        other_user = self.get_superuser()
        link = "/api/users/%s/data-downloads/" % other_user.pk

        response = self.client.get(link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't see other users data downloads."}
        )

    def test_get_empty_list(self):
        """api returns empy list"""
        self.assertFalse(self.user.datadownload_set.exists())

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [])

    def test_populated_list(self):
        """api returns list"""
        for _ in range(6):
            request_user_data_download(self.user)
        self.assertTrue(self.user.datadownload_set.exists())

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 5)



# Generated by Django 3.2.15 on 2023-01-04 12:36

from django.db import migrations
from django.utils import timezone

CHUNK_SIZE = 50


def create_settings(apps, _):
    Subject = apps.get_model("misago_oauth2", "Subject")
    User = apps.get_model("misago_users", "User")

    tz_now = timezone.now()
    sso_users = User.objects.filter(sso_id__isnull=False)

    batch = []

    for user in sso_users.iterator(chunk_size=CHUNK_SIZE):
        batch.append(
            Subject(
                sub=str(user.sso_id),
                user=user,
                created_on=tz_now,
                last_used_on=tz_now,
            )
        )

        if len(batch) >= CHUNK_SIZE:
            Subject.objects.bulk_create(batch)
            batch = []

    if batch:
        Subject.objects.bulk_create(batch)


class Migration(migrations.Migration):
    dependencies = [
        ("misago_oauth2", "0001_initial"),
        ("misago_users", "0022_deleteduser"),
    ]

    operations = [migrations.RunPython(create_settings)]



from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import pgettext

from ...admin.views import render
from ...admin.views.generic import AdminView
from ..models import Setting
from .forms import (
    AnalyticsSettingsForm,
    CaptchaSettingsForm,
    GeneralSettingsForm,
    NotificationsSettingsForm,
    OAuth2SettingsForm,
    ThreadsSettingsForm,
    UsersSettingsForm,
)


def index(request):
    return render(request, "misago/admin/conf/index.html")


class SettingsView(AdminView):
    root_link = None  # Unused by change config views
    template_name = None
    form_class = None

    def get_template_name(self, request):
        return self.template_name

    def dispatch(self, request, *args, **kwargs):
        settings = self.get_settings(self.form_class.settings)
        initial = self.get_initial_form_data(settings)
        form = self.form_class(request=request, initial=initial)
        if request.method == "POST":
            form = self.form_class(
                request.POST, request.FILES, request=request, initial=initial
            )
            if form.is_valid():
                form.save(settings)
                messages.success(
                    request, pgettext("admin settings", "Settings have been saved.")
                )
                return redirect(request.path_info)
        return self.render(request, {"form": form, "form_settings": settings})

    def get_settings(self, form_settings):
        settings = {}
        for setting in Setting.objects.filter(setting__in=form_settings):
            settings[setting.setting] = setting

        if len(settings) != len(form_settings):
            not_found_settings = list(
                set(settings.keys()).symmetric_difference(set(form_settings))
            )
            raise ValueError(
                "Some of settings defined in form could not be found: %s"
                % (", ".join(not_found_settings))
            )

        return settings

    def get_initial_form_data(self, settings):
        return {key: setting.value for key, setting in settings.items()}


class AnalyticsSettingsView(SettingsView):
    form_class = AnalyticsSettingsForm
    template_name = "misago/admin/conf/analytics_settings.html"


class CaptchaSettingsView(SettingsView):
    form_class = CaptchaSettingsForm
    template_name = "misago/admin/conf/captcha_settings.html"


class GeneralSettingsView(SettingsView):
    form_class = GeneralSettingsForm
    template_name = "misago/admin/conf/general_settings.html"


class NotificationsSettingsView(SettingsView):
    form_class = NotificationsSettingsForm
    template_name = "misago/admin/conf/notifications_settings.html"


class OAuth2SettingsView(SettingsView):
    form_class = OAuth2SettingsForm
    template_name = "misago/admin/conf/oauth2_settings.html"


class ThreadsSettingsView(SettingsView):
    form_class = ThreadsSettingsForm
    template_name = "misago/admin/conf/threads_settings.html"


class UsersSettingsView(SettingsView):
    form_class = UsersSettingsForm
    template_name = "misago/admin/conf/users_settings.html"



from datetime import timedelta

import pytest
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ...threads.test import reply_thread
from ..categories import get_categories_new_posts
from ..poststracker import save_read


def remove_tracking(thread):
    thread.started_on = timezone.now() - timedelta(days=4)
    thread.save()
    thread.first_post.posted_on = thread.started_on
    thread.first_post.save()


@pytest.fixture
def read_thread(user, thread):
    save_read(user, thread.first_post)
    return thread


def test_get_categories_new_posts_returns_empty_dict_for_empty_list(request_mock):
    assert get_categories_new_posts(request_mock, []) == {}


def test_get_categories_new_posts_returns_false_for_empty_category(
    request_mock, default_category
):
    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_true_for_new_post(
    request_mock, post, default_category
):
    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: True}


def test_get_categories_new_posts_returns_false_for_post_older_than_user(
    request_mock, post, default_category
):
    post.posted_on = timezone.now() - timedelta(days=1)
    post.save()

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


@override_dynamic_settings(readtracker_cutoff=3)
def test_get_categories_new_posts_returns_false_for_post_older_than_cutoff(
    request_mock, user, post, default_category
):
    user.joined_on = timezone.now() - timedelta(days=5)
    user.save()

    post.posted_on = timezone.now() - timedelta(days=4)
    post.save()

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_false_for_read_post(
    request_mock, user, post, default_category
):
    save_read(user, post)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_true_for_new_post_in_untracked_thread(
    request_mock, thread, default_category
):
    remove_tracking(thread)
    reply_thread(thread)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: True}


def test_get_categories_new_posts_returns_false_for_new_post_in_hidden_thread(
    request_mock, hidden_thread, default_category
):
    reply_thread(hidden_thread)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_false_for_read_post_in_untracked_thread(
    request_mock, user, thread, default_category
):
    remove_tracking(thread)
    post = reply_thread(thread)
    save_read(user, post)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_false_for_read_post_in_read_thread(
    request_mock, user, read_thread, default_category
):
    post = reply_thread(read_thread)
    save_read(user, post)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_false_for_unapproved_post_in_read_thread(
    request_mock, user, read_thread, default_category
):
    reply_thread(read_thread, is_unapproved=True)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_true_for_new_event_in_read_thread(
    request_mock, read_thread, default_category
):
    reply_thread(read_thread, is_event=True)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: True}


def test_get_categories_new_posts_returns_false_for_hidden_event_in_read_thread(
    request_mock, read_thread, default_category
):
    reply_thread(read_thread, is_hidden=True, is_event=True)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_true_for_visible_hidden_event_in_read_thread(
    request_mock, read_thread, default_category
):
    request_mock.user_acl["categories"][default_category.id]["can_hide_events"] = 1
    reply_thread(read_thread, is_hidden=True, is_event=True)

    categories_new_posts = get_categories_new_posts(request_mock, [default_category])
    assert categories_new_posts == {default_category.pk: True}


def test_get_categories_new_posts_returns_false_for_empty_category_for_anonymous_user(
    anonymous_request_mock, default_category
):
    categories_new_posts = get_categories_new_posts(
        anonymous_request_mock, [default_category]
    )
    assert categories_new_posts == {default_category.pk: False}


def test_get_categories_new_posts_returns_false_for_tracked_thread_for_anonymous_user(
    anonymous_request_mock, thread, default_category
):
    categories_new_posts = get_categories_new_posts(
        anonymous_request_mock, [default_category]
    )
    assert categories_new_posts == {default_category.pk: False}


@override_dynamic_settings(readtracker_cutoff=3)
def test_get_categories_new_posts_returns_false_for_untracked_thread_for_anonymous_user(
    anonymous_request_mock, thread, default_category
):
    remove_tracking(thread)

    categories_new_posts = get_categories_new_posts(
        anonymous_request_mock, [default_category]
    )
    assert categories_new_posts == {default_category.pk: False}



from django.core.cache import cache

from . import ACL_CACHE
from ..cache.versions import invalidate_cache


def get_acl_cache(user, cache_versions):
    key = get_cache_key(user, cache_versions)
    return cache.get(key)


def set_acl_cache(user, cache_versions, user_acl):
    key = get_cache_key(user, cache_versions)
    cache.set(key, user_acl)


def get_cache_key(user, cache_versions):
    return "acl_%s_%s" % (user.acl_key, cache_versions[ACL_CACHE])


def clear_acl_cache():
    invalidate_cache(ACL_CACHE)



from datetime import timedelta

from django.contrib.auth import get_user_model
from django.db.models import Q
from django.dispatch import Signal, receiver
from django.utils import timezone
from django.utils.translation import pgettext

from ..core.pgutils import chunk_queryset
from .models import AuditTrail
from .profilefields import profilefields

User = get_user_model()

anonymize_user_data = Signal()
archive_user_data = Signal()
delete_user_content = Signal()
remove_old_ips = Signal()
username_changed = Signal()


@receiver(archive_user_data)
def archive_user_details(sender, archive=None, **kwargs):
    archive.add_dict(
        "details",
        {
            pgettext("archived user details", "Username"): sender.username,
            pgettext("archived user details", "E-mail"): sender.email,
            pgettext("archived user details", "Joined on"): sender.joined_on,
            pgettext("archived user details", "Joined from IP"): (
                sender.joined_from_ip or "unavailable"
            ),
        },
    )


@receiver(archive_user_data)
def archive_user_profile_fields(sender, archive=None, **kwargs):
    clean_profile_fields = {}
    for profile_fields_group in profilefields.get_fields_groups():
        for profile_field in profile_fields_group["fields"]:
            if sender.profile_fields.get(profile_field.fieldname):
                field_value = sender.profile_fields[profile_field.fieldname]
                clean_profile_fields[str(profile_field.label)] = field_value

    if clean_profile_fields:
        archive.add_dict("profile_fields", clean_profile_fields)


@receiver(archive_user_data)
def archive_user_avatar(sender, archive=None, **kwargs):
    archive.add_model_file(sender.avatar_tmp, directory="avatar", prefix="tmp")
    archive.add_model_file(sender.avatar_src, directory="avatar", prefix="src")
    for avatar in sender.avatar_set.iterator():
        archive.add_model_file(avatar.image, directory="avatar", prefix=avatar.size)


@receiver(archive_user_data)
def archive_user_audit_trail(sender, archive=None, **kwargs):
    for audit_trail in chunk_queryset(sender.audittrail_set):
        item_name = audit_trail.created_on.strftime("%H%M%S-audit-trail")
        archive.add_text(item_name, audit_trail.ip_address, date=audit_trail.created_on)


@receiver(archive_user_data)
def archive_user_name_history(sender, archive=None, **kwargs):
    for name_change in sender.namechanges.order_by("id").iterator():
        item_name = name_change.changed_on.strftime("%H%M%S-name-change")
        archive.add_dict(
            item_name,
            {
                pgettext(
                    "archived username history", "New username"
                ): name_change.new_username,
                pgettext(
                    "archived username history", "Old username"
                ): name_change.old_username,
            },
            date=name_change.changed_on,
        )


@receiver(username_changed)
def handle_name_change(sender, **kwargs):
    sender.user_renames.update(changed_by_username=sender.username)


@receiver(remove_old_ips)
def remove_old_registrations_ips(sender, *, ip_storage_time, **kwargs):
    datetime_cutoff = timezone.now() - timedelta(days=ip_storage_time)
    ip_is_too_new = Q(joined_on__gt=datetime_cutoff)
    ip_is_already_removed = Q(joined_from_ip__isnull=True)

    queryset = User.objects.exclude(ip_is_too_new | ip_is_already_removed)
    queryset.update(joined_from_ip=None)


@receiver(remove_old_ips)
def remove_old_audit_trails(sender, *, ip_storage_time, **kwargs):
    removal_cutoff = timezone.now() - timedelta(days=ip_storage_time)
    AuditTrail.objects.filter(created_on__lte=removal_cutoff).delete()


@receiver(anonymize_user_data)
def delete_data_downloads(sender, **kwargs):
    for data_download in chunk_queryset(sender.datadownload_set):
        data_download.delete()



import re

from django.utils.translation import pgettext

QUOTE_HEADER_RE = re.compile(
    r"""
<div class="quote-heading" data-noquote="1">(?P<title>.*?)</div>
""".strip(),
    re.IGNORECASE | re.MULTILINE | re.DOTALL,
)

SPOILER_REVEAL_BTN = '<button class="spoiler-reveal" type="button"></button>'


def finalize_markup(post):
    post = QUOTE_HEADER_RE.sub(replace_quote_headers, post)
    post = replace_spoiler_reveal_buttons(post)
    return post


def replace_quote_headers(matchobj):
    title = matchobj.group("title")
    if title:
        quote_title = pgettext("quote title", "%(title)s has written:") % {
            "title": title
        }
    else:
        quote_title = pgettext("quote title", "Quoted message:")
    return '<div class="quote-heading" data-noquote="1">%s</div>' % quote_title


def replace_spoiler_reveal_buttons(post):
    final_btn = SPOILER_REVEAL_BTN.replace(
        "></", ">%s</" % pgettext("spoiler", "Reveal spoiler")
    )
    return post.replace(SPOILER_REVEAL_BTN, final_btn)



from django.contrib import messages
from django.utils.translation import pgettext_lazy

from ...admin.views import generic
from ..models import MenuItem
from ..cache import clear_menus_cache
from .forms import MenuItemForm
from .ordering import get_next_free_order


class MenuItemAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:settings:menu-items:index"
    model = MenuItem
    form_class = MenuItemForm
    templates_dir = "misago/admin/menuitems"
    message_404 = pgettext_lazy(
        "admin menu items", "Requested menu item does not exist."
    )

    def handle_form(self, form, request, target):
        form.save()

        if self.message_submit:
            messages.success(request, self.message_submit % {"item": target})


class MenuItemsList(MenuItemAdmin, generic.ListView):
    ordering = (("order", None),)
    mass_actions = [
        {
            "action": "delete",
            "name": pgettext_lazy("admin menu items", "Delete items"),
            "confirmation": pgettext_lazy(
                "admin menu items", "Are you sure you want to delete those menu items?"
            ),
        }
    ]

    def action_delete(self, request, items):
        items.delete()
        clear_menus_cache()
        messages.success(
            request,
            pgettext_lazy("admin menu items", "Selected menu items have been deleted."),
        )


class NewMenuItem(MenuItemAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy(
        "admin menu items", "New menu item %(item)s has been saved."
    )

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)
        form.instance.order = get_next_free_order()
        form.instance.save()
        clear_menus_cache()


class EditMenuItem(MenuItemAdmin, generic.ModelFormView):
    message_submit = pgettext_lazy(
        "admin menu items", "Menu item %(item)s has been edited."
    )

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)
        form.instance.save()
        clear_menus_cache()


class DeleteMenuItem(MenuItemAdmin, generic.ButtonView):
    def button_action(self, request, target):
        target.delete()
        clear_menus_cache()
        message = pgettext_lazy(
            "admin menu items", "Menu item %(item)s has been deleted."
        )
        messages.success(request, message % {"item": target})


class MoveDownMenuItem(MenuItemAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = MenuItem.objects.filter(order__gt=target.order)
            other_target = other_target.earliest("order")
        except MenuItem.DoesNotExist:
            other_target = None

        if other_target:
            other_target.order, target.order = target.order, other_target.order
            other_target.save(update_fields=["order"])
            target.save(update_fields=["order"])
            clear_menus_cache()

            message = pgettext_lazy(
                "admin menu items", "Menu item %(item)s has been moved after %(other)s."
            )
            targets_names = {"item": target, "other": other_target}
            messages.success(request, message % targets_names)


class MoveUpMenuItem(MenuItemAdmin, generic.ButtonView):
    def button_action(self, request, target):
        try:
            other_target = MenuItem.objects.filter(order__lt=target.order)
            other_target = other_target.latest("order")
        except MenuItem.DoesNotExist:
            other_target = None

        if other_target:
            other_target.order, target.order = target.order, other_target.order
            other_target.save(update_fields=["order"])
            target.save(update_fields=["order"])
            clear_menus_cache()

            message = pgettext_lazy(
                "admin menu items",
                "Menu item %(item)s has been moved before %(other)s.",
            )
            targets_names = {"item": target, "other": other_target}
            messages.success(request, message % targets_names)



import pytest
from django.core.exceptions import PermissionDenied

from ..permissions import allow_use_notifications


def test_allow_use_notifications_permission_check_passes_authenticated(user):
    allow_use_notifications(user)


def test_allow_use_notifications_permission_check_blocks_anonymous_users(
    anonymous_user,
):
    with pytest.raises(PermissionDenied) as excinfo:
        allow_use_notifications(anonymous_user)

    assert "You must be signed in" in str(excinfo)



from pathlib import Path
from tempfile import TemporaryDirectory

from ..discover import read_pip_install_file


def test_empty_pip_install_file_is_read_to_empty_list():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w"):
            pass

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == []


def test_pip_install_file_is_read_without_comments_lines():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.writelines(["# comment\n", "other\n"])

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["other"]


def test_pip_install_file_is_read_without_comments_after_content():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.write("line # comment")

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["line"]


def test_pip_install_file_is_read_lines_trimmed():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.writelines(["line   \n", "    other\n"])

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["line", "other"]


def test_pip_install_file_is_read_without_duplicate_lines():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.writelines(["line\n", "other\n", "line\n"])

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["line", "other"]


def test_pip_install_file_is_read_without_version_data():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.write("line==0.12rc1\n")

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["line"]


def test_pip_install_file_is_read_with_names_converted_to_package_names():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.write("plugin-name\n")

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["plugin_name"]


def test_pip_install_file_is_read_with_blank_lines_skipped():
    with TemporaryDirectory() as plugins_dir:
        pip_install_path = Path(plugins_dir) / "pip-install.txt"
        with open(pip_install_path, "w") as fp:
            fp.write("plugin-name\n\nline")

        file_contents = read_pip_install_file(pip_install_path)
        assert file_contents == ["plugin_name", "line"]



# Generated by Django 4.2.7 on 2023-11-13 19:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("misago_acl", "0004_cache_version"),
        ("misago_threads", "0012_set_dj_partial_indexes"),
    ]

    operations = [
        migrations.RenameIndex(
            model_name="pollvote",
            new_name="misago_thre_poll_id_c3e8fe_idx",
            old_fields=("poll", "voter_name"),
        ),
        migrations.RenameIndex(
            model_name="post",
            new_name="misago_thre_thread__4e5114_idx",
            old_fields=("thread", "id"),
        ),
        migrations.RenameIndex(
            model_name="post",
            new_name="misago_thre_poster__eefeb2_idx",
            old_fields=("poster", "posted_on"),
        ),
        migrations.RenameIndex(
            model_name="post",
            new_name="misago_thre_is_even_d29c90_idx",
            old_fields=("is_event", "is_hidden"),
        ),
        migrations.RenameIndex(
            model_name="subscription",
            new_name="misago_thre_send_em_da8b99_idx",
            old_fields=("send_email", "last_read_on"),
        ),
        migrations.RenameIndex(
            model_name="thread",
            new_name="misago_thre_categor_6cb52c_idx",
            old_fields=("category", "id"),
        ),
        migrations.RenameIndex(
            model_name="thread",
            new_name="misago_thre_categor_7ec8b0_idx",
            old_fields=("category", "last_post_on"),
        ),
        migrations.RenameIndex(
            model_name="thread",
            new_name="misago_thre_categor_c5eaf1_idx",
            old_fields=("category", "replies"),
        ),
        migrations.AlterField(
            model_name="attachmenttype",
            name="limit_downloads_to",
            field=models.ManyToManyField(
                blank=True, related_name="+", to="misago_acl.role"
            ),
        ),
        migrations.AlterField(
            model_name="attachmenttype",
            name="limit_uploads_to",
            field=models.ManyToManyField(
                blank=True, related_name="+", to="misago_acl.role"
            ),
        ),
    ]



from ...users.bans import get_user_ban
from ...users.models import Rank
from ..users import (
    PASSWORD,
    get_fake_admin_activated_user,
    get_fake_banned_user,
    get_fake_deleted_user,
    get_fake_inactive_user,
    get_fake_user,
    get_fake_username,
)


def test_fake_user_can_be_created(db, fake):
    assert get_fake_user(fake)


def test_fake_user_is_created_with_predictable_password(db, fake):
    user = get_fake_user(fake)
    assert user.check_password(PASSWORD)


def test_fake_user_is_created_with_test_avatars(db, fake):
    user = get_fake_user(fake)
    assert user.avatars


def test_new_fake_user_avatars_are_created_for_every_new_user(db, fake):
    user_a = get_fake_user(fake)
    user_b = get_fake_user(fake)
    assert user_a.avatars != user_b.avatars


def test_fake_user_is_created_with_explicit_rank(db, fake):
    rank = Rank.objects.create(name="Test Rank")
    user = get_fake_user(fake, rank)
    assert user.rank is rank


def test_banned_fake_user_can_be_created(db, cache_versions, fake):
    user = get_fake_banned_user(fake)
    assert get_user_ban(user, cache_versions)


def test_inactivate_fake_user_can_be_created(db, fake):
    user = get_fake_inactive_user(fake)
    assert user.requires_activation


def test_admin_activated_fake_user_can_be_created(db, fake):
    user = get_fake_admin_activated_user(fake)
    assert user.requires_activation


def test_deleted_fake_user_can_be_created(db, fake):
    user = get_fake_deleted_user(fake)
    assert not user.is_active


def test_fake_username_can_be_created(fake):
    assert get_fake_username(fake)


def test_different_fake_username_is_used_every_time(fake):
    fake_usernames = [get_fake_username(fake) for i in range(5)]
    assert len(fake_usernames) == len(set(fake_usernames))



from ...test import assert_contains


def test_favicon_html_is_present_in_html_if_set(client, favicon):
    response = client.get("/")
    assert_contains(response, favicon.image.url)


def test_favicon_32_html_is_present_in_html_if_set(client, favicon_32):
    response = client.get("/")
    assert_contains(response, favicon_32.image.url)


def test_favicon_16_html_is_present_in_html_if_set(client, favicon_16):
    response = client.get("/")
    assert_contains(response, favicon_16.image.url)


def test_apple_touch_icon_html_is_present_in_html_if_set(client, apple_touch_icon):
    response = client.get("/")
    assert_contains(response, apple_touch_icon.image.url)



from django.core.cache import cache

from . import THEME_CACHE
from ..cache.versions import invalidate_cache


def get_theme_cache(cache_versions):
    key = get_cache_key(cache_versions)
    return cache.get(key)


def set_theme_cache(cache_versions, theme):
    key = get_cache_key(cache_versions)
    cache.set(key, theme)


def get_cache_key(cache_versions):
    return "%s_%s" % (THEME_CACHE, cache_versions[THEME_CACHE])


def clear_theme_cache():
    invalidate_cache(THEME_CACHE)



import random
import time

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress
from ....notifications.models import Notification
from ....notifications.verbs import NotificationVerb
from ....threads.models import Post

User = get_user_model()


class Command(BaseCommand):
    help = "Adds notifications entries for all users and posts"

    def handle(self, *args, **options):
        total_users = User.objects.count()
        total_posts = Post.objects.count()

        self.stdout.write(
            f"Creating notifications entries for {total_users} users "
            f"and {total_posts} posts...\n"
        )

        Notification.objects.all().delete()

        total_notifications = total_users * total_posts
        processed_count = 0

        start_time = time.time()
        show_progress(self, processed_count, total_notifications)

        post_fields = (
            "id",
            "thread_id",
            "category_id",
            "thread__title",
            "poster_id",
            "poster_name",
            "posted_on",
        )

        for user in User.objects.iterator():
            for post in Post.objects.values(*post_fields).iterator():
                Notification.objects.create(
                    user=user,
                    category_id=post["category_id"],
                    thread_id=post["thread_id"],
                    thread_title=post["thread__title"],
                    post_id=post["id"],
                    actor_id=post["poster_id"],
                    actor_name=post["poster_name"],
                    verb=NotificationVerb.REPLIED,
                    is_read=random.choice((True, False)),
                    created_at=post["posted_on"],
                )

                processed_count += 1
                show_progress(self, processed_count, total_notifications)

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\nSuccessfully created %s notifications in %s"
        self.stdout.write(message % (total_notifications, total_humanized))



# Generated by Django 3.2.15 on 2023-03-30 22:39

from django.db import migrations
from django.utils import timezone

from ...core.pgutils import chunk_queryset

BATCH_SIZE_LIMIT = 50


def migrate_threads_subscriptions(apps, schema_editor):
    WatchedThread = apps.get_model("misago_notifications", "WatchedThread")
    Subscription = apps.get_model("misago_threads", "Subscription")

    WatchedThread.objects.all().delete()

    timezone_now = timezone.now()
    batch = []
    batch_size = 0

    for subscription in chunk_queryset(Subscription.objects.all()):
        batch.append(
            WatchedThread(
                user_id=subscription.user_id,
                category_id=subscription.category_id,
                thread_id=subscription.thread_id,
                send_emails=subscription.send_email,
                created_at=timezone_now,
                read_at=subscription.last_read_on,
            )
        )
        batch_size += 1

        if batch_size == BATCH_SIZE_LIMIT:
            WatchedThread.objects.bulk_create(batch)
            batch = []
            batch_size = 0

    if batch:
        WatchedThread.objects.bulk_create(batch)


class Migration(migrations.Migration):
    atomic = False
    dependencies = [
        ("misago_notifications", "0001_initial"),
        ("misago_threads", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(migrate_threads_subscriptions, migrations.RunPython.noop),
    ]



from functools import wraps
from typing import Dict, List, Protocol

from django.http import HttpRequest
from django.template import Context
from django.utils.safestring import SafeString, mark_safe

from .enums import PluginOutlet
from .hooks import ActionHook


class PluginOutletHookAction(Protocol):
    def __call__(
        self, request: HttpRequest, context: dict | Context
    ) -> str | SafeString | None:
        pass


class PluginOutletHook(ActionHook[PluginOutletHookAction]):
    __slots__ = ActionHook.__slots__

    def __call__(
        self, request: HttpRequest, context: dict | Context
    ) -> List[str | SafeString | None]:
        return super().__call__(request, context)


template_outlets: Dict[str, PluginOutletHook] = {}


def create_new_outlet(outlet_name: str):
    if outlet_name in template_outlets:
        raise ValueError(f"Template outlet '{outlet_name}' already exists.")

    template_outlets[plugin_outlet.name] = PluginOutletHook()


for plugin_outlet in PluginOutlet:
    create_new_outlet(plugin_outlet.name)


def append_outlet_action(
    outlet_name: str | PluginOutlet, action: PluginOutletHookAction
):
    get_outlet(outlet_name).append_action(action)


def prepend_outlet_action(
    outlet_name: str | PluginOutlet, action: PluginOutletHookAction
):
    get_outlet(outlet_name).prepend_action(action)


def get_outlet(outlet_name: str | PluginOutlet) -> PluginOutletHook:
    try:
        if isinstance(outlet_name, PluginOutlet):
            return template_outlets[outlet_name.name]
        return template_outlets[outlet_name]
    except KeyError as exc:
        raise KeyError(f"Unknown template outlet: {outlet_name}") from exc


def template_outlet_action(f):
    """Decorator for an outlet action that renders a template with returned context."""

    @wraps(f)
    @mark_safe
    def wrapped_outlet_action(request: HttpRequest, context: Context):
        template_data = f(request, context)
        if template_data is None:
            return ""

        if isinstance(template_data, str):
            return _include_template(template_data, context)

        if isinstance(template_data, tuple):
            template_name, new_context = template_data
            return _include_template(template_name, context, new_context)

        return ""

    return wrapped_outlet_action


def _include_template(
    template_name: str, context: Context, new_context: dict | None = None
):
    """Subset of Django include template tag.

    Works only with Django template engine.
    """
    cache = context.render_context.dicts[0].setdefault("_template_outlet_action", {})
    template = cache.get(template_name)
    if template is None:
        template = context.template.engine.get_template(template_name)
        cache[template_name] = template

    if new_context:
        with context.push(**new_context):
            return template.render(context)

    return template.render(context)



import pytest
from django.core.exceptions import ValidationError

from ..validators import validate_image_square, validate_sluggable


def test_sluggable_validator_raises_error_if_result_slug_will_be_empty():
    validator = validate_sluggable()
    with pytest.raises(ValidationError):
        validator("!#@! !@#@")


def test_sluggable_validator_raises_custom_error_if_result_slug_will_be_empty():
    error_message = "I'm short custom error!"
    validator = validate_sluggable(error_short=error_message)
    with pytest.raises(ValidationError) as e:
        validator("!#@! !@#@")
    assert error_message in str(e.value)


def test_sluggable_validator_raises_error_if_result_slug_will_be_too_long():
    validator = validate_sluggable()
    with pytest.raises(ValidationError):
        validator("a" * 256)


def test_sluggable_validator_raises_custom_error_if_result_slug_will_be_too_long():
    error_message = "I'm long custom error!"
    validator = validate_sluggable(error_long=error_message)
    with pytest.raises(ValidationError) as e:
        validator("a" * 256)
    assert error_message in str(e.value)


def test_square_square_validator_validates_square_image(mocker):
    image = mocker.Mock(width=100, height=100)
    validate_image_square(image)


def test_square_square_validator_raises_error_if_image_is_not_square(mocker):
    image = mocker.Mock(width=100, height=200)
    with pytest.raises(ValidationError):
        validate_image_square(image)



from django import template
from django.template import Context
from django.template.loader import get_template

register = template.Library()


def _render_editor_template(context, editor, tpl):
    c = Context(context)
    c["editor"] = editor

    return get_template(tpl).render(c)


def editor_body(context, editor):
    return _render_editor_template(context, editor, editor.body_template)


register.simple_tag(takes_context=True)(editor_body)


def editor_js(context, editor):
    return _render_editor_template(context, editor, editor.js_template)


register.simple_tag(takes_context=True)(editor_js)



from django.test import TestCase

from .. import test
from ...categories.models import Category
from ..utils import add_categories_to_items, get_thread_id_from_url


class AddCategoriesToItemsTests(TestCase):
    def setUp(self):
        """
        Create categories tree for test cases:

        First category (created by migration)

        Category A
          + Category B
            + Subcategory C
            + Subcategory D

        Category E
          + Subcategory F
        """
        super().setUp()

        self.root = Category.objects.root_category()

        Category(
            name="Category A", slug="category-a", css_class="showing-category-a"
        ).insert_at(self.root, position="last-child", save=True)
        Category(
            name="Category E", slug="category-e", css_class="showing-category-e"
        ).insert_at(self.root, position="last-child", save=True)

        self.root = Category.objects.root_category()

        self.category_a = Category.objects.get(slug="category-a")
        Category(
            name="Category B", slug="category-b", css_class="showing-category-b"
        ).insert_at(self.category_a, position="last-child", save=True)

        self.category_b = Category.objects.get(slug="category-b")
        Category(
            name="Category C", slug="category-c", css_class="showing-category-c"
        ).insert_at(self.category_b, position="last-child", save=True)
        Category(
            name="Category D", slug="category-d", css_class="showing-category-d"
        ).insert_at(self.category_b, position="last-child", save=True)

        self.category_c = Category.objects.get(slug="category-c")
        self.category_d = Category.objects.get(slug="category-d")

        self.category_e = Category.objects.get(slug="category-e")
        Category(
            name="Category F", slug="category-f", css_class="showing-category-f"
        ).insert_at(self.category_e, position="last-child", save=True)

        Category.objects.partial_rebuild(self.root.tree_id)

        self.root = Category.objects.root_category()
        self.category_a = Category.objects.get(slug="category-a")
        self.category_b = Category.objects.get(slug="category-b")
        self.category_c = Category.objects.get(slug="category-c")
        self.category_d = Category.objects.get(slug="category-d")
        self.category_e = Category.objects.get(slug="category-e")
        self.category_f = Category.objects.get(slug="category-f")

        self.categories = list(Category.objects.all_categories(include_root=True))

    def test_root_thread_from_root(self):
        """thread in root category is handled"""
        thread = test.post_thread(category=self.root)
        add_categories_to_items(self.root, self.categories, [thread])

        self.assertEqual(thread.category, self.root)

    def test_root_thread_from_elsewhere(self):
        """thread in root category is handled"""
        thread = test.post_thread(category=self.root)
        add_categories_to_items(self.category_e, self.categories, [thread])

        self.assertEqual(thread.category, self.root)

    def test_direct_child_thread_from_parent(self):
        """thread in direct child category is handled"""
        thread = test.post_thread(category=self.category_e)
        add_categories_to_items(self.root, self.categories, [thread])

        self.assertEqual(thread.category, self.category_e)

    def test_direct_child_thread_from_elsewhere(self):
        """thread in direct child category is handled"""
        thread = test.post_thread(category=self.category_e)
        add_categories_to_items(self.category_b, self.categories, [thread])

        self.assertEqual(thread.category, self.category_e)

    def test_child_thread_from_root(self):
        """thread in child category is handled"""
        thread = test.post_thread(category=self.category_d)
        add_categories_to_items(self.root, self.categories, [thread])

        self.assertEqual(thread.category, self.category_d)

    def test_child_thread_from_parent(self):
        """thread in child category is handled"""
        thread = test.post_thread(category=self.category_d)
        add_categories_to_items(self.category_a, self.categories, [thread])

        self.assertEqual(thread.category, self.category_d)

    def test_child_thread_from_category(self):
        """thread in child category is handled"""
        thread = test.post_thread(category=self.category_d)
        add_categories_to_items(self.category_d, self.categories, [thread])

        self.assertEqual(thread.category, self.category_d)

    def test_child_thread_from_elsewhere(self):
        """thread in child category is handled"""
        thread = test.post_thread(category=self.category_d)
        add_categories_to_items(self.category_f, self.categories, [thread])

        self.assertEqual(thread.category, self.category_d)


class MockRequest:
    def __init__(self, scheme, host, wsgialias=""):
        self.scheme = scheme
        self.host = host

        self.path_info = "/api/threads/123/merge/"
        self.path = "%s%s" % (wsgialias.rstrip("/"), self.path_info)

    def get_host(self):
        return self.host

    def is_secure(self):
        return self.scheme == "https"


class GetThreadIdFromUrlTests(TestCase):
    def test_get_thread_id_from_valid_urls(self):
        """get_thread_id_from_url extracts thread pk from valid urls"""
        TEST_CASES = [
            {
                # perfect match
                "request": MockRequest("https", "testforum.com", "/discuss/"),
                "url": "https://testforum.com/discuss/t/test-thread/123/",
                "pk": 123,
            },
            {
                # we don't validate scheme in case site recently moved to https
                # but user still has old url's saved somewhere
                "request": MockRequest("http", "testforum.com", "/discuss/"),
                "url": "http://testforum.com/discuss/t/test-thread/432/post/12321/",
                "pk": 432,
            },
            {
                # extract thread id from other thread urls
                "request": MockRequest("https", "testforum.com", "/discuss/"),
                "url": "http://testforum.com/discuss/t/test-thread/432/post/12321/",
                "pk": 432,
            },
            {
                # extract thread id from thread page url
                "request": MockRequest("http", "testforum.com", "/discuss/"),
                "url": "http://testforum.com/discuss/t/test-thread/432/123/",
                "pk": 432,
            },
            {
                # extract thread id from thread last post url with relative schema
                "request": MockRequest("http", "testforum.com", "/discuss/"),
                "url": "//testforum.com/discuss/t/test-thread/18/last/",
                "pk": 18,
            },
            {
                # extract thread id from url that lacks scheme
                "request": MockRequest("http", "testforum.com", ""),
                "url": "testforum.com/t/test-thread/12/last/",
                "pk": 12,
            },
            {
                # extract thread id from schemaless thread last post url
                "request": MockRequest("http", "testforum.com", "/discuss/"),
                "url": "testforum.com/discuss/t/test-thread/18/last/",
                "pk": 18,
            },
            {
                # extract thread id from url that lacks scheme and hostname
                "request": MockRequest("http", "testforum.com", ""),
                "url": "/t/test-thread/13/",
                "pk": 13,
            },
            {
                # extract thread id from url that has port name
                "request": MockRequest("http", "127.0.0.1:8000", ""),
                "url": "https://127.0.0.1:8000/t/test-thread/13/",
                "pk": 13,
            },
            {
                # extract thread id from url that isn't trimmed
                "request": MockRequest("http", "127.0.0.1:8000", ""),
                "url": "   /t/test-thread/13/   ",
                "pk": 13,
            },
        ]

        for case in TEST_CASES:
            pk = get_thread_id_from_url(case["request"], case["url"])
            self.assertEqual(
                pk,
                case["pk"],
                "get_thread_id_from_url for %(url)s should return %(pk)s" % case,
            )

    def test_get_thread_id_from_invalid_urls(self):
        TEST_CASES = [
            {
                # lacking wsgi alias
                "request": MockRequest("https", "testforum.com"),
                "url": "http://testforum.com/discuss/t/test-thread-123/",
            },
            {
                # invalid wsgi alias
                "request": MockRequest("https", "testforum.com", "/discuss/"),
                "url": "http://testforum.com/forum/t/test-thread-123/",
            },
            {
                # invalid hostname
                "request": MockRequest("http", "misago-project.org", "/discuss/"),
                "url": "https://testforum.com/discuss/t/test-thread-432/post/12321/",
            },
            {
                # old thread url
                "request": MockRequest("http", "testforum.com"),
                "url": "https://testforum.com/thread/test-123/",
            },
            {
                # dashed thread url
                "request": MockRequest("http", "testforum.com"),
                "url": "https://testforum.com/t/test-thread-123/",
            },
            {
                # non-thread url
                "request": MockRequest("http", "testforum.com"),
                "url": "https://testforum.com/user/user-123/",
            },
            {
                # rubbish url
                "request": MockRequest("http", "testforum.com"),
                "url": "asdsadsasadsaSA&das8as*S(A*sa",
            },
            {
                # blank url
                "request": MockRequest("http", "testforum.com"),
                "url": "/",
            },
            {
                # empty url
                "request": MockRequest("http", "testforum.com"),
                "url": "",
            },
        ]

        for case in TEST_CASES:
            pk = get_thread_id_from_url(case["request"], case["url"])
            self.assertIsNone(
                pk, "get_thread_id_from_url for %s should fail" % case["url"]
            )



# Generated by Django 1.11.16 on 2018-12-26 16:11
from django.db import migrations

from .. import THEME_CACHE
from ...cache.operations import StartCacheVersioning


def create_default_theme(apps, schema_editor):
    Theme = apps.get_model("misago_themes", "Theme")
    Theme.objects.create(
        name="default",
        is_default=True,
        is_active=True,
        lft=1,
        rght=2,
        tree_id=0,
        level=0,
    )


class Migration(migrations.Migration):
    dependencies = [("misago_themes", "0001_initial")]

    operations = [
        StartCacheVersioning(THEME_CACHE),
        migrations.RunPython(create_default_theme),
    ]



# Generated by Django 4.2.7 on 2023-12-19 18:07

from django.db import migrations

from ..enums import CategoryTree, CategoryTreeDeprecated


def change_default_categories_trees_ids(apps, schema_editor):
    # Swaps tree ids for threads and private threads around
    Category = apps.get_model("misago_categories", "Category")
    Category.objects.filter(tree_id=CategoryTreeDeprecated.THREADS).update(
        tree_id=CategoryTree.THREADS
    )
    Category.objects.filter(special_role="private_threads").update(
        tree_id=CategoryTree.PRIVATE_THREADS
    )


class Migration(migrations.Migration):
    dependencies = [
        ("misago_categories", "0011_plugin_data"),
    ]

    operations = [migrations.RunPython(change_default_categories_trees_ids)]



from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

from ...notifications.models import Notification
from .. import test
from .test_threads_api import ThreadsApiTestCase


class PostReadApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.post = test.reply_thread(
            self.thread, poster=self.user, posted_on=timezone.now()
        )

        self.api_link = reverse(
            "misago:api:thread-post-read",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    def test_read_anonymous(self):
        """api validates if reading user is authenticated"""
        self.logout_user()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    def test_read_post(self):
        """api marks post as read"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(self.user.postread_set.count(), 1)
        self.user.postread_set.get(post=self.post)

        # one post read, first post is still unread
        self.assertFalse(response.json()["thread_is_read"])

        # read second post
        response = self.client.post(
            reverse(
                "misago:api:thread-post-read",
                kwargs={"thread_pk": self.thread.pk, "pk": self.thread.first_post.pk},
            )
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(self.user.postread_set.count(), 2)
        self.user.postread_set.get(post=self.thread.first_post)

        # both posts are read
        self.assertTrue(response.json()["thread_is_read"])


def test_read_watched_thread_post_updates_watching_read_at_if_its_older_than_post(
    user, user_client, thread, reply, watched_thread_factory
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)
    watched_thread.read_at = timezone.now() - timedelta(seconds=1)
    watched_thread.save()

    response = user_client.post(
        reverse(
            "misago:api:thread-post-read",
            kwargs={"thread_pk": thread.pk, "pk": reply.pk},
        )
    )
    assert response.status_code == 200

    watched_thread.refresh_from_db()
    assert watched_thread.read_at == reply.posted_on


def test_read_watched_thread_post_skips_watching_read_at_if_its_newer_than_post(
    user, user_client, thread, reply, watched_thread_factory
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)
    watched_thread.read_at = timezone.now() + timedelta(seconds=5)
    watched_thread.save()

    response = user_client.post(
        reverse(
            "misago:api:thread-post-read",
            kwargs={"thread_pk": thread.pk, "pk": reply.pk},
        )
    )
    assert response.status_code == 200

    watched_thread.refresh_from_db()
    assert watched_thread.read_at > reply.posted_on


def test_read_thread_post_reads_its_unread_notifications(
    user, user_client, thread, reply
):
    notification = Notification.objects.create(
        user=user,
        verb="TEST",
        post=reply,
        is_read=False,
    )

    response = user_client.post(
        reverse(
            "misago:api:thread-post-read",
            kwargs={"thread_pk": thread.pk, "pk": reply.pk},
        )
    )
    assert response.status_code == 200

    notification.refresh_from_db()
    assert notification.is_read

    user.refresh_from_db()
    assert user.unread_notifications == 0


def test_read_thread_post_updates_unread_notifications_count(
    user, user_client, thread, reply
):
    user.unread_notifications = 5
    user.save()

    notification = Notification.objects.create(
        user=user,
        verb="TEST",
        post=reply,
        is_read=False,
    )
    other_notification = Notification.objects.create(
        user=user,
        verb="TEST",
        post=reply,
        is_read=False,
    )

    response = user_client.post(
        reverse(
            "misago:api:thread-post-read",
            kwargs={"thread_pk": thread.pk, "pk": reply.pk},
        )
    )
    assert response.status_code == 200

    notification.refresh_from_db()
    assert notification.is_read

    other_notification.refresh_from_db()
    assert other_notification.is_read

    user.refresh_from_db()
    assert user.unread_notifications == 3


def test_read_thread_post_excludes_other_users_notifications(
    user, user_client, other_user, thread, reply
):
    user.unread_notifications = 5
    user.save()

    notification = Notification.objects.create(
        user=user,
        verb="TEST",
        post=reply,
        is_read=False,
    )
    other_user_notification = Notification.objects.create(
        user=other_user,
        verb="TEST",
        post=reply,
        is_read=False,
    )

    response = user_client.post(
        reverse(
            "misago:api:thread-post-read",
            kwargs={"thread_pk": thread.pk, "pk": reply.pk},
        )
    )
    assert response.status_code == 200

    notification.refresh_from_db()
    assert notification.is_read

    other_user_notification.refresh_from_db()
    assert not other_user_notification.is_read

    user.refresh_from_db()
    assert user.unread_notifications == 4



import pytest

from ..models import Notification, WatchedThread


def test_user_delete_deletes_user_notifications(user, other_user):
    notification = Notification.objects.create(
        user=user,
        actor=other_user,
        actor_name=other_user.username,
        verb="TEST",
    )

    user.delete(anonymous_username="Deleted")

    with pytest.raises(Notification.DoesNotExist):
        notification.refresh_from_db()


def test_user_delete_clears_notifications_actor(user, other_user):
    notification = Notification.objects.create(
        user=user,
        actor=other_user,
        actor_name=other_user.username,
        verb="TEST",
    )

    other_user.delete(anonymous_username="Deleted")

    notification.refresh_from_db()
    assert notification.actor is None


def test_user_delete_excludes_other_users_notifications(user, other_user):
    notification = Notification.objects.create(
        user=other_user,
        verb="TEST",
    )

    user.delete(anonymous_username="Deleted")

    notification.refresh_from_db()


def test_user_delete_deletes_users_watched_threads(
    user, thread, watched_thread_factory
):
    watched_thread = watched_thread_factory(user, thread, send_emails=True)

    user.delete(anonymous_username="Deleted")

    with pytest.raises(WatchedThread.DoesNotExist):
        watched_thread.refresh_from_db()


def test_user_delete_excludes_other_users_watched_threads(
    user, other_user, thread, watched_thread_factory
):
    watched_thread = watched_thread_factory(other_user, thread, send_emails=True)

    user.delete(anonymous_username="Deleted")

    watched_thread.refresh_from_db()



from django.urls import path

from ...core.apirouter import MisagoApiRouter
from ..api import auth, captcha, mention
from ..api.ranks import RanksViewSet
from ..api.usernamechanges import UsernameChangesViewSet
from ..api.users import UserViewSet

urlpatterns = [
    path("auth/", auth.gateway, name="auth"),
    path("auth/criteria/", auth.get_criteria, name="auth-criteria"),
    path("auth/send-activation/", auth.send_activation, name="send-activation"),
    path(
        "auth/send-password-form/",
        auth.send_password_form,
        name="send-password-form",
    ),
    path(
        "auth/change-password/<int:pk>/<slug:token>/",
        auth.change_forgotten_password,
        name="change-forgotten-password",
    ),
    path("captcha-question/", captcha.question, name="captcha-question"),
    path("mention/", mention.mention_suggestions, name="mention-suggestions"),
]

router = MisagoApiRouter()
router.register(r"ranks", RanksViewSet)
router.register(r"users", UserViewSet)
router.register(r"username-changes", UsernameChangesViewSet, basename="usernamechange")
urlpatterns += router.urls



from django.db import migrations

_ = lambda s: s


def create_default_categories_roles(apps, schema_editor):
    CategoryRole = apps.get_model("misago_categories", "CategoryRole")

    CategoryRole.objects.create(
        name=_("See only"),
        permissions={
            # categories perms
            "misago.categories.permissions": {"can_see": 1, "can_browse": 0}
        },
    )

    read_only = CategoryRole.objects.create(
        name=_("Read only"),
        permissions={
            # categories perms
            "misago.categories.permissions": {"can_see": 1, "can_browse": 1},
            # threads perms
            "misago.threads.permissions.threads": {
                "can_see_all_threads": 1,
                "can_see_posts_likes": 1,
                "can_download_other_users_attachments": 1,
                "can_like_posts": 1,
            },
        },
    )

    CategoryRole.objects.create(
        name=_("Reply to threads"),
        permissions={
            # categories perms
            "misago.categories.permissions": {"can_see": 1, "can_browse": 1},
            # threads perms
            "misago.threads.permissions.threads": {
                "can_see_all_threads": 1,
                "can_reply_threads": 1,
                "can_edit_posts": 1,
                "can_download_other_users_attachments": 1,
                "max_attachment_size": 500,
                "can_see_posts_likes": 2,
                "can_like_posts": 1,
            },
        },
    )

    standard = CategoryRole.objects.create(
        name=_("Start and reply threads"),
        permissions={
            # categories perms
            "misago.categories.permissions": {"can_see": 1, "can_browse": 1},
            # threads perms
            "misago.threads.permissions.threads": {
                "can_see_all_threads": 1,
                "can_start_threads": 1,
                "can_reply_threads": 1,
                "can_edit_threads": 1,
                "can_edit_posts": 1,
                "can_download_other_users_attachments": 1,
                "max_attachment_size": 500,
                "can_see_posts_likes": 2,
                "can_like_posts": 1,
            },
        },
    )

    moderator = CategoryRole.objects.create(
        name=_("Moderator"),
        permissions={
            # categories perms
            "misago.categories.permissions": {"can_see": 1, "can_browse": 1},
            # threads perms
            "misago.threads.permissions.threads": {
                "can_see_all_threads": 1,
                "can_start_threads": 1,
                "can_reply_threads": 1,
                "can_edit_threads": 2,
                "can_edit_posts": 2,
                "can_hide_own_threads": 2,
                "can_hide_own_posts": 2,
                "thread_edit_time": 0,
                "post_edit_time": 0,
                "can_hide_threads": 2,
                "can_hide_posts": 2,
                "can_protect_posts": 1,
                "can_move_posts": 1,
                "can_merge_posts": 1,
                "can_announce_threads": 1,
                "can_pin_threads": 2,
                "can_close_threads": 1,
                "can_move_threads": 1,
                "can_merge_threads": 1,
                "can_approve_content": 1,
                "can_download_other_users_attachments": 1,
                "max_attachment_size": 2500,
                "can_delete_other_users_attachments": 1,
                "can_see_posts_likes": 2,
                "can_like_posts": 1,
                "can_report_content": 1,
                "can_see_reports": 1,
                "can_hide_events": 2,
            },
        },
    )

    # assign category roles to roles
    Category = apps.get_model("misago_categories", "Category")
    Role = apps.get_model("misago_acl", "Role")
    RoleCategoryACL = apps.get_model("misago_categories", "RoleCategoryACL")

    category = Category.objects.get(tree_id=1, level=1)

    RoleCategoryACL.objects.create(
        role=Role.objects.get(name=_("Moderator")),
        category=category,
        category_role=moderator,
    )

    RoleCategoryACL.objects.create(
        role=Role.objects.get(special_role="authenticated"),
        category=category,
        category_role=standard,
    )

    RoleCategoryACL.objects.create(
        role=Role.objects.get(special_role="anonymous"),
        category=category,
        category_role=read_only,
    )


class Migration(migrations.Migration):
    dependencies = [
        ("misago_categories", "0002_default_categories"),
        ("misago_acl", "0003_default_roles"),
    ]

    operations = [migrations.RunPython(create_default_categories_roles)]



from unittest.mock import Mock, patch

import pytest
from requests.exceptions import Timeout

from ...conf.test import override_dynamic_settings
from .. import exceptions
from ..client import REQUESTS_TIMEOUT, get_access_token

ACCESS_TOKEN = "acc3ss-t0k3n"
CODE_GRANT = "valid-code"


@pytest.fixture
def mock_request(dynamic_settings):
    return Mock(
        settings=dynamic_settings,
        build_absolute_uri=lambda url: f"http://mysite.com{url or ''}",
    )


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
)
def test_access_token_is_retrieved_using_post_request(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value={
                    "access_token": ACCESS_TOKEN,
                },
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_access_token(mock_request, CODE_GRANT) == ACCESS_TOKEN

        post_mock.assert_called_once_with(
            "https://example.com/oauth2/token",
            data={
                "grant_type": "authorization_code",
                "client_id": "clientid123",
                "client_secret": "secr3t",
                "redirect_uri": "http://mysite.com/oauth2/complete/",
                "code": CODE_GRANT,
            },
            headers={},
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
    oauth2_token_extra_headers="Accept: application/json\nApi-Ver:1234",
)
def test_access_token_is_retrieved_using_extra_headers(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value={
                    "access_token": ACCESS_TOKEN,
                },
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_access_token(mock_request, CODE_GRANT) == ACCESS_TOKEN

        post_mock.assert_called_once_with(
            "https://example.com/oauth2/token",
            data={
                "grant_type": "authorization_code",
                "client_id": "clientid123",
                "client_secret": "secr3t",
                "redirect_uri": "http://mysite.com/oauth2/complete/",
                "code": CODE_GRANT,
            },
            headers={
                "Accept": "application/json",
                "Api-Ver": "1234",
            },
            timeout=REQUESTS_TIMEOUT,
        )


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
    oauth2_json_token_path="data.auth.token",
)
def test_access_token_is_extracted_from_json(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value={
                    "data": {
                        "auth": {
                            "token": ACCESS_TOKEN,
                        },
                    },
                },
            ),
        ),
    )

    with patch("requests.post", post_mock):
        assert get_access_token(mock_request, CODE_GRANT) == ACCESS_TOKEN

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
)
def test_exception_is_raised_if_access_token_request_times_out(mock_request):
    post_mock = Mock(side_effect=Timeout())

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2AccessTokenRequestError):
            get_access_token(mock_request, CODE_GRANT)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
)
def test_exception_is_raised_if_access_token_request_response_is_not_200(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=400,
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2AccessTokenResponseError):
            get_access_token(mock_request, CODE_GRANT)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
)
def test_exception_is_raised_if_access_token_request_response_is_not_json(mock_request):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                side_effect=ValueError(),
            ),
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2AccessTokenJSONError):
            get_access_token(mock_request, CODE_GRANT)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
)
def test_exception_is_raised_if_access_token_request_response_json_is_not_object(
    mock_request,
):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value=["json", "list"],
            ),
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2AccessTokenJSONError):
            get_access_token(mock_request, CODE_GRANT)

        post_mock.assert_called_once()


@override_dynamic_settings(
    oauth2_client_id="clientid123",
    oauth2_client_secret="secr3t",
    oauth2_token_url="https://example.com/oauth2/token",
)
def test_exception_is_raised_if_access_token_request_response_json_misses_token(
    mock_request,
):
    post_mock = Mock(
        return_value=Mock(
            status_code=200,
            json=Mock(
                return_value={
                    "no_token": "nope",
                },
            ),
        ),
    )

    with patch("requests.post", post_mock):
        with pytest.raises(exceptions.OAuth2AccessTokenNotProvidedError):
            get_access_token(mock_request, CODE_GRANT)

        post_mock.assert_called_once()



# Generated by Django 1.11.16 on 2018-11-29 20:28
from django.db import migrations, models

from .. import BANS_CACHE


def populate_cache_version(apps, _):
    BanCache = apps.get_model("misago_users", "BanCache")
    BanCache.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [("misago_users", "0016_cache_version")]

    operations = [
        migrations.RemoveField(model_name="bancache", name="bans_version"),
        migrations.RunPython(populate_cache_version, migrations.RunPython.noop),
        migrations.AddField(
            model_name="bancache",
            name="cache_version",
            field=models.CharField(max_length=8),
        ),
    ]



from unittest.mock import Mock

from ..context_processors import conf
from ..test import override_dynamic_settings


def test_request_settings_are_included_in_template_context(db, dynamic_settings):
    mock_request = Mock(settings=dynamic_settings)
    context_settings = conf(mock_request)["settings"]
    assert context_settings == mock_request.settings


def test_settings_are_included_in_frontend_context(db, client):
    response = client.get("/")
    assert response.status_code == 200
    assert '"SETTINGS": {"' in response.content.decode("utf-8")


def test_admin_panel_link_is_included_in_frontend_context_for_admins(admin_client):
    response = admin_client.get("/")
    assert response.status_code == 200
    assert '"ADMIN_URL": "' in response.content.decode("utf-8")


@override_dynamic_settings(show_admin_panel_link_in_ui=False)
def test_admin_panel_link_is_included_from_frontend_context_for_admins_if_disabled(
    admin_client,
):
    response = admin_client.get("/")
    assert response.status_code == 200
    assert '"ADMIN_URL": "' not in response.content.decode("utf-8")


def test_admin_panel_link_is_excluded_from_frontend_context_for_users(user_client):
    response = user_client.get("/")
    assert response.status_code == 200
    assert '"ADMIN_URL": "' not in response.content.decode("utf-8")


def test_admin_panel_link_is_excluded_from_frontend_context_for_guests(db, client):
    response = client.get("/")
    assert response.status_code == 200
    assert '"ADMIN_URL": "' not in response.content.decode("utf-8")



from django.urls import path
from django.utils.translation import pgettext_lazy

from .views import plugins_list


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        urlpatterns.namespace("plugins/", "plugins")

        urlpatterns.patterns("plugins", path("", plugins_list, name="index"))

    def register_navigation_nodes(self, site):
        site.add_node(
            name=pgettext_lazy("admin node", "Plugins"),
            icon="fa fa-cube",
            after="settings:index",
            namespace="plugins",
        )



from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ... import SOCIALAUTH_CACHE


def test_top_provider_can_be_moved_down(admin_client, provider, other_provider):
    provider.order = 0
    provider.save()

    other_provider.order = 1
    other_provider.save()

    admin_client.post(
        reverse("misago:admin:settings:socialauth:down", kwargs={"pk": provider.pk})
    )

    provider.refresh_from_db()
    assert provider.order == 1
    other_provider.refresh_from_db()
    assert other_provider.order == 0


def test_top_provider_cant_be_moved_up(admin_client, provider, other_provider):
    provider.order = 0
    provider.save()

    other_provider.order = 1
    other_provider.save()

    admin_client.post(
        reverse("misago:admin:settings:socialauth:up", kwargs={"pk": provider.pk})
    )

    provider.refresh_from_db()
    assert provider.order == 0
    other_provider.refresh_from_db()
    assert other_provider.order == 1


def test_bottom_provider_cant_be_moved_down(admin_client, provider, other_provider):
    provider.order = 1
    provider.save()

    other_provider.order = 0
    other_provider.save()

    admin_client.post(
        reverse("misago:admin:settings:socialauth:down", kwargs={"pk": provider.pk})
    )

    provider.refresh_from_db()
    assert provider.order == 1
    other_provider.refresh_from_db()
    assert other_provider.order == 0


def test_bottom_provider_can_be_moved_up(admin_client, provider, other_provider):
    provider.order = 1
    provider.save()

    other_provider.order = 0
    other_provider.save()

    admin_client.post(
        reverse("misago:admin:settings:socialauth:up", kwargs={"pk": provider.pk})
    )

    provider.refresh_from_db()
    assert provider.order == 0
    other_provider.refresh_from_db()
    assert other_provider.order == 1


def test_moving_provider_down_invalidates_providers_cache(
    admin_client, provider, other_provider
):
    provider.order = 0
    provider.save()

    other_provider.order = 1
    other_provider.save()

    with assert_invalidates_cache(SOCIALAUTH_CACHE):
        admin_client.post(
            reverse("misago:admin:settings:socialauth:down", kwargs={"pk": provider.pk})
        )


def test_moving_provider_up_invalidates_providers_cache(
    admin_client, provider, other_provider
):
    provider.order = 1
    provider.save()

    other_provider.order = 0
    other_provider.save()

    with assert_invalidates_cache(SOCIALAUTH_CACHE):
        admin_client.post(
            reverse("misago:admin:settings:socialauth:up", kwargs={"pk": provider.pk})
        )


def test_disabled_provider_cant_be_moved_down(
    admin_client, disabled_provider, provider
):
    provider.order = 1
    provider.save()

    disabled_provider.order = 0
    disabled_provider.save()

    admin_client.post(
        reverse(
            "misago:admin:settings:socialauth:down", kwargs={"pk": disabled_provider.pk}
        )
    )

    disabled_provider.refresh_from_db()
    assert disabled_provider.order == 0
    provider.refresh_from_db()
    assert provider.order == 1


def test_disabled_provider_cant_be_moved_up(admin_client, disabled_provider, provider):
    provider.order = 0
    provider.save()

    disabled_provider.order = 1
    disabled_provider.save()

    admin_client.post(
        reverse(
            "misago:admin:settings:socialauth:up", kwargs={"pk": disabled_provider.pk}
        )
    )

    disabled_provider.refresh_from_db()
    assert disabled_provider.order == 1
    provider.refresh_from_db()
    assert provider.order == 0


def test_moving_up_not_setup_provider_has_no_errors(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:socialauth:up", kwargs={"pk": "facebook"})
    )
    assert response.status_code == 302


def test_moving_down_not_setup_provider_has_no_errors(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:socialauth:down", kwargs={"pk": "facebook"})
    )
    assert response.status_code == 302


def test_moving_up_undefined_provider_has_no_errors(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:socialauth:up", kwargs={"pk": "undefined"})
    )
    assert response.status_code == 302


def test_moving_down_undefined_provider_has_no_errors(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:socialauth:down", kwargs={"pk": "undefined"})
    )
    assert response.status_code == 302



from rest_framework.response import Response

from ....notifications.models import Notification, WatchedThread
from ....readtracker import poststracker, threadstracker
from ....readtracker.signals import thread_read
from ....users.models import User
from ...models import Post


def post_read_endpoint(request, thread, watched_thread, post):
    poststracker.make_read_aware(request, post)
    if post.is_new:
        poststracker.save_read(request.user, post)

        update_watched_thread(watched_thread, post)
        read_post_notifications(request.user, post)

    threadstracker.make_read_aware(request, thread)

    # send signal if post read marked thread as read
    # used in some places, eg. syncing unread thread count
    if post.is_new and thread.is_read:
        thread_read.send(request.user, thread=thread)

    return Response({"thread_is_read": thread.is_read})


def update_watched_thread(watched_thread: WatchedThread | None, post: Post):
    if watched_thread and watched_thread.read_at < post.posted_on:
        watched_thread.read_at = post.posted_on
        watched_thread.save(update_fields=["read_at"])


def read_post_notifications(user: User, post: Post):
    updated_notifications = Notification.objects.filter(
        user=user, post=post, is_read=False
    ).update(is_read=True)
    if updated_notifications:
        new_unread_notifications = max(
            [0, user.unread_notifications - updated_notifications]
        )

        if user.unread_notifications != new_unread_notifications:
            user.unread_notifications = new_unread_notifications
            user.save(update_fields=["unread_notifications"])



import json
import os
from pathlib import Path

from ...acl.test import patch_user_acl
from ...conf import settings
from ...conf.test import override_dynamic_settings
from ..avatars import gallery, store
from ..models import AvatarGallery
from ..test import AuthenticatedUserTestCase, create_test_user

TESTFILES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testfiles")
TEST_AVATAR_PATH = os.path.join(TESTFILES_DIR, "avatar.png")


class UserAvatarTests(AuthenticatedUserTestCase):
    """tests for user avatar RPC (/api/users/1/avatar/)"""

    def setUp(self):
        super().setUp()
        self.link = "/api/users/%s/avatar/" % self.user.pk
        self.client.post(self.link, data={"avatar": "generated"})

    def get_current_user(self):
        self.user.refresh_from_db()
        return self.user

    def assertOldAvatarsAreDeleted(self, user):
        self.assertEqual(user.avatar_set.count(), len(settings.MISAGO_AVATARS_SIZES))

    @override_dynamic_settings(allow_custom_avatars=False)
    def test_avatars_off(self):
        """custom avatars are not allowed"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertTrue(options["generated"])
        self.assertFalse(options["gravatar"])
        self.assertFalse(options["crop_src"])
        self.assertFalse(options["crop_tmp"])
        self.assertFalse(options["upload"])
        self.assertFalse(options["galleries"])

    @override_dynamic_settings(allow_custom_avatars=True)
    def test_avatars_on(self):
        """custom avatars are allowed"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertTrue(options["generated"])
        self.assertTrue(options["gravatar"])
        self.assertFalse(options["crop_src"])
        self.assertFalse(options["crop_tmp"])
        self.assertTrue(options["upload"])
        self.assertFalse(options["galleries"])

    def test_gallery_exists(self):
        """api returns gallery"""
        gallery.load_avatar_galleries()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertTrue(options["galleries"])

    def test_avatar_locked(self):
        """requests to api error if user's avatar is locked"""
        self.user.is_avatar_locked = True
        self.user.avatar_lock_user_message = "Your avatar is pwnt."
        self.user.save()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {
                "detail": "Your avatar is locked. You can't change it.",
                "reason": "<p>Your avatar is pwnt.</p>",
            },
        )

    def test_other_user_avatar(self):
        """requests to api error if user tries to access other user"""
        self.logout_user()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have to sign in to perform this action."}
        )

        self.login_user(create_test_user("Other_User", "otheruser@example.com"))

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't change other users avatars."}
        )

    def test_empty_requests(self):
        """empty request errors with code 400"""
        response = self.client.post(self.link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Unknown avatar type."})

    def test_failed_gravatar_request(self):
        """no gravatar RPC fails"""
        self.user.email_hash = "wolololo"
        self.user.save()

        response = self.client.post(self.link, data={"avatar": "gravatar"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "No Gravatar is associated with your e-mail address."},
        )

    def test_successful_gravatar_request(self):
        """gravatar RPC passes"""
        self.user.set_email("rafio.xudb@gmail.com")
        self.user.save()

        response = self.client.post(self.link, data={"avatar": "gravatar"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json()["detail"], "Gravatar was downloaded and set as new avatar."
        )

        self.assertOldAvatarsAreDeleted(self.user)

    def test_generation_request(self):
        """generated avatar is set"""
        response = self.client.post(self.link, data={"avatar": "generated"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json()["detail"], "New avatar based on your account was set."
        )

        self.assertOldAvatarsAreDeleted(self.user)

    def test_avatar_upload_and_crop(self):
        """avatar can be uploaded and cropped"""
        response = self.client.post(self.link, data={"avatar": "upload"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "No file was sent."})

        with open(TEST_AVATAR_PATH, "rb") as avatar:
            response = self.client.post(
                self.link, data={"avatar": "upload", "image": avatar}
            )
            self.assertEqual(response.status_code, 200)

            response_json = response.json()
            self.assertTrue(response_json["crop_tmp"])
            self.assertEqual(
                self.get_current_user().avatar_tmp.url, response_json["crop_tmp"]["url"]
            )

        avatar = Path(self.get_current_user().avatar_tmp.path)
        self.assertTrue(avatar.exists())
        self.assertTrue(avatar.is_file())

        response = self.client.post(
            self.link,
            json.dumps(
                {"avatar": "crop_tmp", "crop": {"offset": {"x": 0, "y": 0}, "zoom": 1}}
            ),
            content_type="application/json",
        )

        response_json = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["detail"], "Uploaded avatar was set.")

        self.assertFalse(self.get_current_user().avatar_tmp)
        self.assertOldAvatarsAreDeleted(self.user)

        avatar = Path(self.get_current_user().avatar_src.path)
        self.assertTrue(avatar.exists())
        self.assertTrue(avatar.is_file())

        response = self.client.post(
            self.link,
            json.dumps(
                {"avatar": "crop_tmp", "crop": {"offset": {"x": 0, "y": 0}, "zoom": 1}}
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "This avatar type is not available."}
        )

        response = self.client.post(
            self.link,
            json.dumps(
                {"avatar": "crop_src", "crop": {"offset": {"x": 0, "y": 0}, "zoom": 1}}
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["detail"], "Avatar was re-cropped.")
        self.assertOldAvatarsAreDeleted(self.user)

        # delete user avatars, test if it deletes src and tmp
        store.delete_avatar(self.get_current_user())

        self.assertTrue(self.get_current_user().avatar_src.path)

        avatar = Path(self.get_current_user().avatar_src.path)
        self.assertFalse(avatar.exists())
        self.assertFalse(avatar.is_file())

    def test_gallery_set_empty_gallery(self):
        """gallery handles set avatar on empty gallery"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            self.link, data={"avatar": "galleries", "image": 123}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "This avatar type is not available."}
        )

    def test_gallery_image_validation(self):
        """gallery validates image to set"""
        gallery.load_avatar_galleries()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        # no image id is handled
        response = self.client.post(self.link, data={"avatar": "galleries"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Incorrect image."})

        # invalid id is handled
        response = self.client.post(
            self.link, data={"avatar": "galleries", "image": "asdsadsadsa"}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Incorrect image."})

        # nonexistant image is handled
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertTrue(options["galleries"])

        test_avatar = options["galleries"][0]["images"][0]["id"]
        response = self.client.post(
            self.link, data={"avatar": "galleries", "image": test_avatar + 5000}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Incorrect image."})

        # default gallery image is handled
        AvatarGallery.objects.filter(pk=test_avatar).update(
            gallery=gallery.DEFAULT_GALLERY
        )

        response = self.client.post(
            self.link, data={"avatar": "galleries", "image": test_avatar}
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Incorrect image."})

    def test_gallery_set_valid_avatar(self):
        """its possible to set avatar from gallery"""
        gallery.load_avatar_galleries()

        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertTrue(options["galleries"])

        test_avatar = options["galleries"][0]["images"][0]["id"]
        response = self.client.post(
            self.link, data={"avatar": "galleries", "image": test_avatar}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["detail"], "Avatar from gallery was set.")
        self.assertOldAvatarsAreDeleted(self.user)


class UserAvatarModerationTests(AuthenticatedUserTestCase):
    """tests for moderate user avatar RPC (/api/users/1/moderate-avatar/)"""

    def setUp(self):
        super().setUp()

        self.other_user = create_test_user("Other_User", "other@user.com")
        self.link = "/api/users/%s/moderate-avatar/" % self.other_user.pk

    @patch_user_acl({"can_moderate_avatars": 0})
    def test_no_permission(self):
        """no permission to moderate avatar"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't moderate avatars."})

    @patch_user_acl({"can_moderate_avatars": 1})
    def test_moderate_avatar(self):
        """moderate avatar"""
        response = self.client.get(self.link)
        self.assertEqual(response.status_code, 200)

        options = response.json()
        self.assertEqual(options["is_avatar_locked"], self.other_user.is_avatar_locked)
        self.assertEqual(
            options["avatar_lock_user_message"],
            self.other_user.avatar_lock_user_message,
        )
        self.assertEqual(
            options["avatar_lock_staff_message"],
            self.other_user.avatar_lock_staff_message,
        )

        response = self.client.post(
            self.link,
            json.dumps(
                {
                    "is_avatar_locked": True,
                    "avatar_lock_user_message": "Test user message.",
                    "avatar_lock_staff_message": "Test staff message.",
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        self.other_user.refresh_from_db()

        options = response.json()
        self.assertEqual(self.other_user.is_avatar_locked, True)
        self.assertEqual(self.other_user.avatar_lock_user_message, "Test user message.")
        self.assertEqual(
            self.other_user.avatar_lock_staff_message, "Test staff message."
        )

        self.assertEqual(options["avatars"], self.other_user.avatars)
        self.assertEqual(options["is_avatar_locked"], self.other_user.is_avatar_locked)
        self.assertEqual(
            options["avatar_lock_user_message"],
            self.other_user.avatar_lock_user_message,
        )
        self.assertEqual(
            options["avatar_lock_staff_message"],
            self.other_user.avatar_lock_staff_message,
        )

        response = self.client.post(
            self.link,
            json.dumps(
                {
                    "is_avatar_locked": False,
                    "avatar_lock_user_message": None,
                    "avatar_lock_staff_message": None,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        self.other_user.refresh_from_db()
        self.assertFalse(self.other_user.is_avatar_locked)
        self.assertIsNone(self.other_user.avatar_lock_user_message)
        self.assertIsNone(self.other_user.avatar_lock_staff_message)

        options = response.json()
        self.assertEqual(options["avatars"], self.other_user.avatars)
        self.assertEqual(options["is_avatar_locked"], self.other_user.is_avatar_locked)
        self.assertEqual(
            options["avatar_lock_user_message"],
            self.other_user.avatar_lock_user_message,
        )
        self.assertEqual(
            options["avatar_lock_staff_message"],
            self.other_user.avatar_lock_staff_message,
        )

        response = self.client.post(
            self.link,
            json.dumps(
                {
                    "is_avatar_locked": True,
                    "avatar_lock_user_message": "",
                    "avatar_lock_staff_message": "",
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.is_avatar_locked)
        self.assertEqual(self.other_user.avatar_lock_user_message, "")
        self.assertEqual(self.other_user.avatar_lock_staff_message, "")

        options = response.json()
        self.assertEqual(options["avatars"], self.other_user.avatars)
        self.assertEqual(options["is_avatar_locked"], self.other_user.is_avatar_locked)
        self.assertEqual(
            options["avatar_lock_user_message"],
            self.other_user.avatar_lock_user_message,
        )
        self.assertEqual(
            options["avatar_lock_staff_message"],
            self.other_user.avatar_lock_staff_message,
        )

        response = self.client.post(
            self.link,
            json.dumps({"is_avatar_locked": False}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        self.other_user.refresh_from_db()
        self.assertFalse(self.other_user.is_avatar_locked)
        self.assertEqual(self.other_user.avatar_lock_user_message, "")
        self.assertEqual(self.other_user.avatar_lock_staff_message, "")

        options = response.json()
        self.assertEqual(options["avatars"], self.other_user.avatars)
        self.assertEqual(options["is_avatar_locked"], self.other_user.is_avatar_locked)
        self.assertEqual(
            options["avatar_lock_user_message"],
            self.other_user.avatar_lock_user_message,
        )
        self.assertEqual(
            options["avatar_lock_staff_message"],
            self.other_user.avatar_lock_staff_message,
        )

    @patch_user_acl({"can_moderate_avatars": 1})
    def test_moderate_own_avatar(self):
        """moderate own avatar"""
        response = self.client.get("/api/users/%s/moderate-avatar/" % self.user.pk)
        self.assertEqual(response.status_code, 200)



from unittest.mock import Mock

from ....notifications.models import Notification
from ..pagination import paginate_queryset


def test_pagination_returns_all_items(notifications):
    request = Mock(GET={})
    page = paginate_queryset(request, Notification.objects, "id", 20)

    assert len(page.items) == 15
    assert not page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == [f"test_{i}" for i in range(15)]
    assert page.first_cursor == page.items[0].id
    assert page.last_cursor == page.items[-1].id


def test_pagination_returns_no_items(db):
    request = Mock(GET={})
    page = paginate_queryset(request, Notification.objects, "id", 20)

    assert len(page.items) == 0
    assert not page.has_next
    assert not page.has_previous
    assert page.first_cursor is None
    assert page.last_cursor is None


def test_pagination_returns_firsts_and_last_cursor(notifications):
    request = Mock(GET={})
    page = paginate_queryset(request, Notification.objects, "id", 20)

    assert len(page.items) == 15
    assert page.first_cursor == page.items[0].id
    assert page.last_cursor == page.items[-1].id


def test_pagination_returns_all_items_in_reverse_order(notifications):
    request = Mock(GET={})
    page = paginate_queryset(request, Notification.objects, "-id", 20)

    assert len(page.items) == 15
    assert not page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == [f"test_{i}" for i in reversed(range(15))]


def test_pagination_returns_first_items_up_to_max_limit(notifications):
    request = Mock(GET={})
    page = paginate_queryset(request, Notification.objects, "id", 4)

    assert len(page.items) == 4
    assert page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_0", "test_1", "test_2", "test_3"]


def test_pagination_returns_first_items_up_to_given_limit(notifications):
    request = Mock(GET={"limit": 5})
    page = paginate_queryset(request, Notification.objects, "id", 10)

    assert len(page.items) == 5
    assert page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_0", "test_1", "test_2", "test_3", "test_4"]


def test_pagination_returns_first_items_up_to_max_limit_in_reverse_order(
    notifications,
):
    request = Mock(GET={})
    page = paginate_queryset(request, Notification.objects, "-id", 4)

    assert len(page.items) == 4
    assert page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_14", "test_13", "test_12", "test_11"]


def test_pagination_returns_first_items_up_to_given_limit_in_reverse_order(
    notifications,
):
    request = Mock(GET={"limit": 5})
    page = paginate_queryset(request, Notification.objects, "-id", 10)

    assert len(page.items) == 5
    assert page.has_next
    assert not page.has_previous

    items_verbs = [item.verb for item in page.items]
    assert items_verbs == ["test_14", "test_13", "test_12", "test_11", "test_10"]



from django.urls import reverse

from ...test import assert_contains
from ...threads.models import ThreadParticipant
from ..verbs import NotificationVerb
from ..models import Notification
from ..registry import registry


def test_notification_view_returns_redirect_to_user_notification(rf, user, user_client):
    request = rf.get("/notification/1/")
    notification = Notification.objects.create(user=user, verb="TEST")
    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302
    assert response.headers["location"] == registry.get_redirect_url(
        request, notification
    )


def test_notification_view_sets_notification_as_read(user, user_client):
    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)
    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302

    notification.refresh_from_db()
    assert notification.is_read


def test_notification_view_updates_user_unread_notifications_count(user, user_client):
    user.unread_notifications = 1
    user.save()

    notification = Notification.objects.create(user=user, verb="TEST", is_read=False)
    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302

    user.refresh_from_db()
    assert user.unread_notifications == 0


def test_notification_view_skips_user_unread_notifications_count_for_read_notification(
    user, user_client
):
    user.unread_notifications = 1
    user.save()

    notification = Notification.objects.create(user=user, verb="TEST", is_read=True)
    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302

    user.refresh_from_db()
    assert user.unread_notifications == 1


def test_notification_view_returns_404_error_for_nonexisting_notification(user_client):
    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": 1})
    )
    assert response.status_code == 404


def test_notification_view_returns_404_error_for_other_user_notification(
    other_user, user_client
):
    notification = Notification.objects.create(user=other_user, verb="TEST")
    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 404


def test_notification_view_shows_permission_denied_page_to_guests(db, client):
    response = client.get(reverse("misago:notification", kwargs={"notification_id": 1}))
    assert_contains(response, "You must be signed in", status_code=403)


def test_notification_view_returns_redirect_to_thread_reply(
    user, user_client, default_category, thread, reply
):
    notification = Notification.objects.create(
        user=user,
        verb=NotificationVerb.REPLIED,
        category=default_category,
        thread=thread,
        thread_title=thread.title,
        post=reply,
    )

    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302
    assert response.headers["location"]


def test_notification_view_returns_redirect_to_private_thread_reply(
    user, user_client, private_threads_category, private_thread, private_thread_reply
):
    ThreadParticipant.objects.create(thread=private_thread, user=user)

    notification = Notification.objects.create(
        user=user,
        verb=NotificationVerb.REPLIED,
        category=private_threads_category,
        thread=private_thread,
        thread_title=private_thread.title,
        post=private_thread_reply,
    )

    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302
    assert response.headers["location"]


def test_notification_view_returns_redirect_to_private_thread_invite(
    user, user_client, private_threads_category, private_thread
):
    ThreadParticipant.objects.create(thread=private_thread, user=user)

    notification = Notification.objects.create(
        user=user,
        verb=NotificationVerb.INVITED,
        category=private_threads_category,
        thread=private_thread,
        thread_title=private_thread.title,
        post_id=private_thread.first_post_id,
    )

    response = user_client.get(
        reverse("misago:notification", kwargs={"notification_id": notification.id})
    )
    assert response.status_code == 302
    assert response.headers["location"]



from rest_framework import serializers


def test_post_validator(context, data):
    title_match = "casino" in data.get("title", "").lower()
    post_match = "casino" in data.get("post", "").lower()

    if title_match or post_match:
        raise serializers.ValidationError("Don't discuss gambling!")



from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from ..management.commands import listusedprofilefields
from ..test import create_test_user


class ListUsedProfileFieldsTests(TestCase):
    def test_no_fields_set(self):
        """utility has no showstoppers when no fields are set"""
        create_test_user("User", "user@example.com")

        out = StringIO()
        call_command(listusedprofilefields.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "No profile fields are currently in use.")

    def test_fields_set(self):
        """utility lists number of users that have different fields set"""
        create_test_user(
            "User1",
            "user1@example.com",
            profile_fields={"gender": "male", "bio": "Yup!"},
        )
        create_test_user(
            "User2", "user2@example.com", profile_fields={"gender": "male"}
        )
        create_test_user("User3", "user3@example.com", profile_fields={"location": ""})

        out = StringIO()
        call_command(listusedprofilefields.Command(), stdout=out)
        command_output = [l.strip() for l in out.getvalue().strip().splitlines()]

        self.assertEqual(command_output, ["bio:      1", "gender:   2", "location: 1"])



from datetime import timedelta
from io import StringIO

from django.core import management
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..management.commands import clearreadtracker
from ..models import PostRead


def call_command():
    command = clearreadtracker.Command()

    out = StringIO()
    management.call_command(command, stdout=out)
    return out.getvalue().strip().splitlines()[-1].strip()


def test_command_works_if_there_are_no_read_tracker_entries(db):
    command_output = call_command()
    assert command_output == "No expired entries were found"


@override_dynamic_settings(readtracker_cutoff=5)
def test_recent_read_tracker_entry_is_not_cleared(user, post):
    existing = PostRead.objects.create(
        user=user,
        category=post.category,
        thread=post.thread,
        post=post,
        last_read_on=timezone.now(),
    )

    command_output = call_command()
    assert command_output == "No expired entries were found"
    assert PostRead.objects.exists()


@override_dynamic_settings(readtracker_cutoff=5)
def test_old_read_tracker_entry_is_cleared(user, post):
    existing = PostRead.objects.create(
        user=user,
        category=post.category,
        thread=post.thread,
        post=post,
        last_read_on=timezone.now() - timedelta(days=10),
    )

    command_output = call_command()
    assert command_output == "Deleted 1 expired entries"
    assert not PostRead.objects.exists()



from ..threads.models import Post
from ..threads.permissions import exclude_invisible_posts
from .cutoffdate import get_cutoff_date


def make_read_aware(request, threads):
    if not threads:
        return

    if not hasattr(threads, "__iter__"):
        threads = [threads]

    make_read(threads)

    if request.user.is_anonymous:
        return

    categories = [t.category for t in threads]
    cutoff_date = get_cutoff_date(request.settings, request.user)

    queryset = (
        Post.objects.filter(thread__in=threads, posted_on__gt=cutoff_date)
        .values_list("thread", flat=True)
        .distinct()
    )

    queryset = queryset.exclude(id__in=request.user.postread_set.values("post"))
    queryset = exclude_invisible_posts(request.user_acl, categories, queryset)

    unread_threads = list(queryset)

    for thread in threads:
        if thread.pk in unread_threads:
            thread.is_read = False
            thread.is_new = True


def make_read(threads):
    for thread in threads:
        thread.is_read = True
        thread.is_new = False



from django.contrib import messages
from django.db import transaction
from django.shortcuts import redirect

from .base import AdminView


class TargetedView(AdminView):
    is_atomic = True

    def dispatch(self, request, *args, **kwargs):
        if self.is_atomic:
            with transaction.atomic():
                return self.wrapped_dispatch(request, *args, **kwargs)
        return self.wrapped_dispatch(request, *args, **kwargs)

    def wrapped_dispatch(self, request, *args, **kwargs):
        target = self.get_target_or_none(request, kwargs)
        if not target:
            messages.error(request, self.message_404)
            return redirect(self.root_link)

        error = self.check_permissions(  # pylint: disable=assignment-from-no-return
            request, target
        )
        if error:
            messages.error(request, error)
            return redirect(self.root_link)

        return self.real_dispatch(request, target)

    def get_target_or_none(self, request, kwargs):
        try:
            return self.get_target(kwargs)
        except self.get_model().DoesNotExist:
            return None

    def get_target(self, kwargs):
        if len(kwargs) > 1:
            raise ValueError("TargetedView.get_target() received more than one kwarg")
        if len(kwargs) != 1:
            return self.get_model()()

        queryset = self.get_model().objects
        if self.is_atomic:
            queryset = queryset.select_for_update()
        (pk,) = kwargs.values()
        return queryset.get(pk=pk)

    def check_permissions(self, request, target):
        pass

    def real_dispatch(self, request, target):
        pass


class FormView(TargetedView):
    form_class = None
    template_name = "form.html"

    def get_form_class(self, request):
        return self.form_class

    def get_form(self, form_class, request):
        if request.method == "POST":
            return form_class(request.POST, request.FILES)
        return form_class()

    def handle_form(self, form, request):
        raise NotImplementedError(
            "You have to define your own handle_form method to handle form submissions."
        )

    def real_dispatch(self, request, target):
        FormType = self.get_form_class(request)
        form = self.get_form(FormType, request)

        if request.method == "POST" and form.is_valid():
            response = self.handle_form(form, request)

            if response:
                return response
            if "stay" in request.POST:
                return redirect(request.path)
            return redirect(self.root_link)

        return self.render(request, {"form": form})


class ModelFormView(FormView):
    message_submit = None

    def get_form_class(self, request, target):
        return self.form_class

    def get_form(self, form_class, request, target):
        if request.method == "POST":
            return form_class(request.POST, request.FILES, instance=target)
        return form_class(instance=target)

    def handle_form(self, form, request, target):
        form.instance.save()
        if self.message_submit:
            messages.success(request, self.message_submit % {"name": target.name})

    def real_dispatch(self, request, target):
        form_class = self.get_form_class(request, target)
        form = self.get_form(form_class, request, target)

        if request.method == "POST" and form.is_valid():
            response = self.handle_form(  # pylint: disable=assignment-from-no-return
                form, request, target
            )
            if response:
                return response
            if "stay" in request.POST:
                return redirect(request.path)
            return redirect(self.root_link)

        template_name = self.get_template_name(request, target)
        return self.render(request, {"form": form, "target": target}, template_name)

    def get_template_name(self, request, target):
        return "%s/%s" % (self.templates_dir, self.template_name)


class ButtonView(TargetedView):
    def real_dispatch(self, request, target):
        if request.method == "POST":
            new_response = self.button_action(request, target)
            if new_response:
                return new_response
        return redirect(self.root_link)

    def button_action(self, request, target):
        raise NotImplementedError("You have to define custom button_action.")
